from __future__ import annotations

from typing import TYPE_CHECKING

from defusedxml import ElementTree as ET

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import export
from dissect.target.plugins.apps.chat.chat import (
    ChatAttachmentRecord,
    ChatMessageRecord,
    ChatPlugin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.plugins.general.users import UserDetails
    from dissect.target.target import Target


class MSNPlugin(ChatPlugin):
    """Microsoft MSN Messenger plugin.

    Supports the following versions on Windows XP and Windows 7:
        - Windows Live Messenger (WLM) 2009
        - MSN 7.5

    Other versions might work but have not been tested. Does not support ``Messenger Plus! Live`` artifacts.
    Tested using Escargot (https://escargot.chat).

    References:
        - https://en.wikipedia.org/wiki/Microsoft_Messenger_service
        - https://en.wikipedia.org/wiki/MSN_Messenger
        - http://computerforensics.parsonage.co.uk/downloads/MSNandLiveMessengerArtefactsOfConversations.pdf
    """

    __namespace__ = "msn"

    DATA_PATH = "Application Data\\Microsoft\\MSN Messenger"
    HIST_PATH = "My Documents\\My Received Files"

    def __init__(self, target: Target):
        super().__init__(target)
        self.installs = list(self.find_installs())

    def find_installs(self) -> Iterator[tuple[UserDetails, TargetPath]]:
        for user_details in self.target.user_details.all_with_home():
            if (path := self.target.fs.path(user_details.user.home).joinpath(self.DATA_PATH)).exists():
                for profile in path.iterdir():
                    if profile.is_dir():
                        yield user_details, profile

    def check_compatible(self) -> None:
        if not self.installs:
            raise UnsupportedPluginError("No Microsoft MSN installs found on target")

    @export(record=[ChatMessageRecord, ChatAttachmentRecord])
    def history(self) -> Iterator[ChatMessageRecord | ChatAttachmentRecord]:
        """Yield MSN chat history messages.

        Chat history artifacts can be found in:
            - ``$HOME/My Documents/My Received Files/MsnMsgr.txt``
            - ``$HOME/My Documents/My Received Files/$username$PassportID/History/*.xml``
        """

        for user_details, profile in self.installs:
            if not (hist_root := self.target.fs.path(user_details.user.home).joinpath(self.HIST_PATH)).exists():
                self.target.log.warning(
                    "User %s does not have saved MSN chat history: directory %s does not exist",
                    user_details.user.name,
                    hist_root,
                )
                continue

            hist_dir = None
            for item in hist_root.iterdir():
                if item.is_dir() and (hist_dir := item.name).endswith(profile.name):
                    for hist_file in hist_root.joinpath(hist_dir).joinpath("History").glob("*.xml"):
                        try:
                            xml = ET.fromstring(hist_file.read_text())
                        except Exception as e:
                            self.target.log.warning("XML file %s is malformed: %s", hist_file, e)
                            continue

                        for entry in xml:
                            common = {
                                "ts": entry.attrib.get("DateTime", 0),
                                "client": self.__namespace__,
                                "account": profile.name,
                                "sender": entry.find(".//From/User").get("FriendlyName"),
                                "_user": user_details.user,
                                "_target": self.target,
                            }

                            if entry.tag == "Message":
                                yield ChatMessageRecord(
                                    **common,
                                    recipient=entry.find(".//To/User").get("FriendlyName"),
                                    message=entry.find(".//Text").text,
                                )

                            elif (
                                entry.tag in ["Invitation", "InvitationResponse"]
                                and (file := entry.find(".//File")) is not None
                            ):
                                yield ChatAttachmentRecord(
                                    **common,
                                    recipient=None,  # unknown with Invitations
                                    attachment=file.text,
                                    description=entry.find(".//Text").text,
                                )


def convert_email(string: str) -> int:
    """Convert MSN email address to 10 digit Passport ID."""
    num = 0
    for char in string.lower():
        num = num * 101 + ord(char)
        num -= (num // 4294967296) * 4294967296
    return num



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.linux.proc import ProcPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_cmdline(target_linux_users: Target, fs_linux_proc: VirtualFilesystem) -> None:
    target_linux_users.add_plugin(ProcPlugin)
    results = list(target_linux_users.cmdline())
    assert len(results) == 4



#!/usr/bin/env python
from __future__ import annotations

import argparse
import bz2
import dataclasses
import datetime
import enum
import functools
import gzip
import itertools
import json
import sys
from collections import deque
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO, TextIO

import structlog
from flow.record import Record, RecordDescriptor, RecordStreamWriter
from flow.record.adapter.jsonfile import JsonfileWriter
from flow.record.jsonpacker import JsonRecordPacker

try:
    import lz4.frame

    HAS_LZ4 = True
except ImportError:
    HAS_LZ4 = False


try:
    if sys.version_info >= (3, 14):
        from compression import zstd  # novermin
    else:
        from backports import zstd

    HAS_ZSTD = True
except ImportError:
    HAS_ZSTD = False

from dissect.target.exceptions import FatalError, PluginNotFoundError, UnsupportedPluginError
from dissect.target.target import Target
from dissect.target.tools.utils.cli import (
    configure_generic_arguments,
    configure_plugin_arguments,
    execute_function_on_target,
    find_and_filter_plugins,
    open_targets,
    process_generic_arguments,
    process_plugin_arguments,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator

    from flow.record import Record
    from typing_extensions import Self

    from dissect.target.helpers.record import TargetRecordDescriptor
    from dissect.target.plugin import FunctionDescriptor

log = structlog.get_logger(__name__)


@dataclass
class RecordStreamElement:
    target: Target
    func: FunctionDescriptor
    record: Record
    end_pos: int | None = None
    sink_path: Path | None = None


def get_targets(targets: list[str]) -> Iterator[Target]:
    """Return a generator with :class:`Target` objects for provided paths."""
    yield from Target.open_all(targets)


def execute_function(
    target: Target, function: FunctionDescriptor, dry_run: bool, arguments: list[str]
) -> Iterator[TargetRecordDescriptor]:
    """Execute function ``function`` on provided target ``target`` and return a generator
    with the records produced.

    Only output type ``record`` is supported for plugin functions.
    """

    local_log = log.bind(func=function, target=target)
    local_log.debug("Function execution")

    if dry_run:
        print(f"  execute: {function.name} ({function.path})")
        return

    if function.output != "record":
        local_log.info(
            "Skipping target/func pair since its output type is not a record",
            target=target.path,
            func=function.name,
        )
        return
    try:
        _, target_attr = execute_function_on_target(target, function)
    except UnsupportedPluginError as e:
        local_log.error("Unsupported plugin for %s: %s", function.name, e.root_cause_str())
        local_log.debug("%s", function, exc_info=e)
        return
    except PluginNotFoundError:
        local_log.error("Cannot find plugin `%s`", function)
        return
    except FatalError as e:
        e.emit_last_message(local_log.error)
        sys.exit(1)
    except Exception as e:
        local_log.error("Exception while executing function %s (%s): %s", function.name, function.path, e)
        local_log.debug("", exc_info=e)
        local_log.debug("Function info: %s", function)
        return

    # no support for function-specific arguments
    result = target_attr() if callable(target_attr) else target_attr

    try:
        yield from result
    except Exception as e:
        local_log.error("Error occurred while processing output of %s.%s: %s", function.qualname, function.name, e)
        local_log.debug("", exc_info=e)
        return


def produce_target_func_pairs(
    targets: Iterable[Target],
    state: DumpState,
) -> Iterator[tuple[Target, FunctionDescriptor]]:
    """Return a generator with target and function pairs for execution.

    Target and function pairs that correspond to finished sinks in provided state ``state`` are skipped.
    """
    pairs_to_skip = set()
    if state:
        pairs_to_skip.update((str(sink.target_path), sink.func) for sink in state.finished_sinks)

    for target in targets:
        for func_def in find_and_filter_plugins(state.functions, target, state.excluded_functions):
            if state and (target.path, func_def.name) in pairs_to_skip:
                log.info(
                    "Skipping target/func pair since its marked as done in provided state",
                    target=target.path,
                    func=func_def.name,
                    state=state.path,
                )
                continue

            yield (target, func_def)
            state.mark_as_finished(target, func_def.name)


def execute_functions(
    target_func_stream: Iterable[tuple[Target, FunctionDescriptor]], dry_run: bool, arguments: list[str]
) -> Iterator[RecordStreamElement]:
    """Execute a function on a target for target / function pairs in the stream.

    Returns a generator of ``RecordStreamElement`` objects.
    """
    for target, func in target_func_stream:
        for record in execute_function(target, func, dry_run, arguments):
            yield RecordStreamElement(target=target, func=func, record=record)


def log_progress(stream: Iterable[Any], step_size: int = 1000) -> Iterator[Any]:
    """Log a number of items that went though the generator stream after
    every N element (N is configured in ``step_size``).
    """
    i = 0
    targets = set()
    sinks = set()
    for i, element in enumerate(stream, start=1):
        yield element

        if i % step_size == 0:
            log.info("Processing status", processed=i, last=element)

        targets.add(element.target)
        if element.sink_path:
            sinks.add(element.sink_path)

    log.info("Pipeline stats", elements=i, targets=len(targets), sinks=len(sinks))


def sink_records(
    record_stream: Iterable[RecordStreamElement],
    state: DumpState,
) -> Iterator[RecordStreamElement]:
    """Persist records from the stream into appropriate sinks, per serialization, compression and record type."""
    with cached_sink_writers(state) as write_element:
        for element in record_stream:
            write_element(element)
            yield element


def persist_processing_state(
    record_stream: Iterable[RecordStreamElement],
    state: DumpState,
) -> Iterator[RecordStreamElement]:
    """Keep track of the pipeline state in a persistent state object."""
    with persisted_state(state) as save_state_periodically:
        for element in record_stream:
            save_state_periodically()
            yield element


def configure_state(args: argparse.Namespace) -> DumpState | None:
    state = None if args.restart else load_state(output_dir=args.output)

    if state:
        log.info("Resuming state", output=args.output, state=state.path)

        if (
            state.serialization != args.serialization
            or state.compression != args.compression
            or state.functions != args.function
            or state.excluded_functions != args.excluded_functions
            or state.target_paths != args.targets
        ):
            log.error(
                "Configuration of the existing state conflicts with parameters provided",
                state_serialization=state.serialization,
                state_compression=state.compression,
                state_targets=state.target_paths,
                state_functions=state.functions,
                state_excluded_fucntions=state.excluded_functions,
                targets=args.targets,
                functions=args.function,
                serialization=args.serialization,
                compression=args.compression,
                state_path=state.path,
            )
            return None
    else:
        state = create_state(
            output_dir=args.output,
            target_paths=args.targets,
            functions=args.function,
            excluded_functions=args.excluded_functions,
            serialization=args.serialization,
            compression=args.compression,
        )
        log.info("New state created", restart=args.restart, state=state.path)

    return state


STATE_FILE_NAME = "target-dump.state.json"

PENDING_UPDATES_LIMIT = 10


@dataclass
class Sink:
    target_path: str
    func: str
    path: Path
    is_dirty: bool = True
    record_count: int = 0
    size_bytes: int = 0

    def __post_init__(self):
        self.func = getattr(self.func, "name", self.func)


@dataclass
class DumpState:
    target_paths: list[str]
    functions: str
    excluded_functions: list[str]
    serialization: str
    compression: str
    start_time: datetime.datetime
    last_update_time: datetime.datetime

    sinks: list[Sink] = dataclasses.field(default_factory=list)

    # Volatile properties
    output_dir: Path | None = None
    pending_updates_count: int | None = 0

    @property
    def record_count(self) -> int:
        return sum(s.record_count for s in self.sinks)

    @property
    def finished_sinks(self) -> list[Sink]:
        return [sink for sink in self.sinks if not sink.is_dirty]

    @property
    def path(self) -> Path:
        return DumpState.get_state_path(self.output_dir)

    @classmethod
    def get_state_path(cls, output_dir: Path) -> Path:
        return output_dir / STATE_FILE_NAME

    def get_full_sink_path(self, sink: Sink) -> Path:
        if not self.output_dir:
            raise ValueError("Output directory is unknown for the state")
        return self.output_dir / sink.path

    def get_sink(self, path: Path) -> Sink | None:
        for sink in self.sinks:
            if sink.path == path:
                return sink
        return None

    def serialize(self) -> str:
        """Serialize state instance into a JSON formatted string."""
        state_dict = dataclasses.asdict(self)
        state_dict.pop("output_dir")
        state_dict.pop("pending_updates_count")
        return json.dumps(
            state_dict,
            default=serialize_obj,
            indent=4,
            sort_keys=True,
        )

    def persist(self, fh: TextIO) -> None:
        """Write serialized state instance into profided ``fh`` byte stream, overwriting it from the beginning."""
        fh.seek(0)
        fh.write(self.serialize())
        fh.flush()
        self.pending_updates_count = 0
        log.debug("State flushed")

    def mark_as_finished(self, target: Target, func: str) -> None:
        """Mark sinks that match provided ``target`` and ``func`` pair as not dirty."""
        matching_sinks = [
            sink for sink in self.sinks if str(sink.target_path) == str(target.path) and sink.func == func
        ]
        for sink in matching_sinks:
            sink.is_dirty = False

    def create_sink(self, sink_path: Path, stream_element: RecordStreamElement) -> Sink:
        """Create a sink instance for provided ``sink_path`` and ``stream_element``
        (from which ``target`` and ``func`` properties are used).
        """
        sink = Sink(
            path=sink_path,
            target_path=str(stream_element.target.path),
            func=stream_element.func,
        )
        self.sinks.append(sink)
        return sink

    def update(self, stream_element: RecordStreamElement, fp_position: int) -> None:
        """Update a sink instance for provided ``stream_element``."""
        sink = self.get_sink(stream_element.sink_path)

        sink.record_count += 1
        sink.size_bytes = fp_position
        self.last_update_time = get_current_utc_time()

        self.pending_updates_count += 1
        log.debug("State updated", records=self.record_count, sinks=len(self.sinks), path=self.path)

    @classmethod
    def from_dict(cls, state_dict: dict) -> Self:
        """Deserialize state instance from provided dictionary."""
        return cls(
            target_paths=state_dict["target_paths"],
            functions=state_dict["functions"],
            excluded_functions=state_dict["excluded_functions"],
            serialization=Serialization(state_dict["serialization"]),
            compression=Compression(state_dict["compression"]),
            start_time=parse_datetime_iso(state_dict["start_time"]),
            last_update_time=parse_datetime_iso(state_dict["last_update_time"]),
            sinks=[
                Sink(
                    target_path=sink["target_path"],
                    path=Path(sink["path"]),
                    func=sink["func"],
                    record_count=sink["record_count"],
                    size_bytes=sink["size_bytes"],
                    is_dirty=sink["is_dirty"],
                )
                for sink in state_dict["sinks"]
            ],
        )

    @classmethod
    def from_path(cls, output_dir: Path) -> Self | None:
        """Deserialize state instance from a file in the provided output directory path."""
        state_path = cls.get_state_path(output_dir)
        if not state_path.exists():
            return None

        with state_path.open(mode="r") as fh:
            try:
                state_dict = json.load(fh)
            except ValueError as e:
                log.warning("Can not load state from path", path=state_path, exc=e)
                return None

        state = cls.from_dict(state_dict)
        state.output_dir = output_dir
        return state

    def get_invalid_sinks(self) -> list[Sink]:
        """Return sinks that have a mismatch between recorded size and a real file size."""
        invalid_sinks = []
        for sink in self.sinks:
            # sink file does not exist
            if not self.get_full_sink_path(sink).exists():
                invalid_sinks.append(sink)
                continue

            # recorded file size for a clean sink is incorrect
            if not sink.is_dirty and sink.size_bytes != sink.path.stat().st_size:
                invalid_sinks.append(sink)
                continue

        return invalid_sinks

    def drop_invalid_sinks(self) -> None:
        """Remove sinks that have a mismatch between recorded size and a real file size from the list of sinks."""
        for invalid_sink in self.get_invalid_sinks():
            self.sinks.remove(invalid_sink)
            log.debug("Ignoring invalid sink", sink=invalid_sink.path)

    def drop_dirty_sinks(self) -> None:
        """Drop sinks that are marked as "dirty" in the current state from the list of sinks."""
        dirty_sinks = [s for s in self.sinks if s.is_dirty]
        for dirty_sink in dirty_sinks:
            self.sinks.remove(dirty_sink)
            log.debug("Ignoring dirty sink", sink=dirty_sink.path)


def create_state(
    *,
    output_dir: Path,
    target_paths: list[str],
    functions: str,
    excluded_functions: list[str],
    serialization: Serialization,
    compression: Compression = None,
) -> DumpState:
    """Create a ``DumpState`` instance with provided properties."""
    current_time = get_current_utc_time()
    return DumpState(
        target_paths=target_paths,
        functions=functions,
        excluded_functions=excluded_functions,
        serialization=serialization,
        compression=compression,
        start_time=current_time,
        last_update_time=current_time,
        output_dir=output_dir,
    )


@contextmanager
def persisted_state(state: DumpState) -> Iterator[Callable]:
    """Return a context manager for persisting ``DumpState`` instance."""

    def save_state(fh: TextIO) -> None:
        state.persist(fh)
        state.pending_updates_count = 0

    def save_state_when_over_limit(fh: TextIO) -> None:
        if state.pending_updates_count >= PENDING_UPDATES_LIMIT:
            save_state(fh)

    state.path.parent.mkdir(parents=True, exist_ok=True)

    with state.path.open(mode="w") as fh:
        try:
            yield functools.partial(save_state_when_over_limit, fh)
        finally:
            save_state(fh)


def load_state(output_dir: Path) -> DumpState | None:
    """Load persisted ``DumpState`` instance from provided ``output_dir`` path and perform sink validation."""
    state = DumpState.from_path(output_dir)

    if not state:
        return None

    # Dropping sinks that are marked as finished (clean) but the file
    # size on the disk is different from the one stored in the state object.
    state.drop_invalid_sinks()

    # Since there are no guarantees about the order of produced records
    # and the records themselves do not have unique IDs, we can not easily verify
    # what records are already stored in the dirty sink.
    # It is easier to just restart all dirty sinks.
    state.drop_dirty_sinks()

    return state


def serialize_obj(obj: Any) -> str:
    """JSON serializer for object types not serializable by ``json`` library."""
    if isinstance(obj, datetime.datetime):
        return obj.isoformat()
    if isinstance(obj, Path):
        return str(obj)
    if isinstance(obj, enum.Enum):
        return obj.value
    raise TypeError(f"Type {type(obj)} not serializable")


class Compression(str, enum.Enum):
    """Supported compression types."""

    BZIP2 = "bzip2"
    GZIP = "gzip"
    LZ4 = "lz4"
    ZSTD = "zstandard"
    NONE = None


class Serialization(str, enum.Enum):
    """Supported serialization methods."""

    JSONLINES = "jsonlines"
    MSGPACK = "msgpack"


COMPRESSION_TO_EXT = {
    Compression.BZIP2: "bz2",
    Compression.GZIP: "gz",
    Compression.LZ4: "lz4",
    Compression.ZSTD: "zstd",
    Compression.NONE: "",
}


DEST_DIR_CACHE_SIZE = 10
DEST_FILENAME_CACHE_SIZE = 10
OPEN_WRITERS_LIMIT = 10


def get_nested_attr(obj: Any, nested_attr: str) -> Any:
    parts = nested_attr.split(".")
    return functools.reduce(getattr, [obj, *parts])


@functools.lru_cache(maxsize=DEST_DIR_CACHE_SIZE)
def get_sink_dir_by_target(target: Target, function: FunctionDescriptor) -> Path:
    func_first_name, _, _ = function.name.partition(".")
    return Path(target.name) / func_first_name


@functools.lru_cache(maxsize=DEST_DIR_CACHE_SIZE)
def get_sink_dir_by_func(target: Target, function: FunctionDescriptor) -> Path:
    func_first_name, _, _ = function.name.partition(".")
    return Path(func_first_name) / target.name


def slugify_descriptor_name(descriptor_name: str) -> str:
    return descriptor_name.replace("/", "_")


@functools.lru_cache(maxsize=DEST_FILENAME_CACHE_SIZE)
def get_sink_filename(
    record_descriptor: RecordDescriptor,
    serialization: Serialization,
    compression: Compression | None = None,
) -> str:
    """Return a sink filename for provided record descriptor, serialization and compression."""
    record_type = slugify_descriptor_name(record_descriptor.name)

    serialization_details = SERIALIZERS[serialization]
    serialization_ext = serialization_details["ext"]

    parts = [record_type, serialization_ext]

    compression_ext = COMPRESSION_TO_EXT[compression]
    if compression_ext:
        parts.append(compression_ext)

    return ".".join(parts)


def get_relative_sink_path(
    element: RecordStreamElement, serialization: str, compression: Compression | None = None
) -> Path:
    """Return a sink path relative to an output directory."""
    sink_dir = get_sink_dir_by_target(element.target, element.func)
    sink_filename = get_sink_filename(element.record._desc, serialization, compression)
    return sink_dir / sink_filename


def open_path(path: Path, mode: str, compression: Compression | None = None) -> BinaryIO:
    """Open ``path`` using ``mode``, with specified ``compression`` and return a file object."""
    if compression == Compression.GZIP:
        fh = gzip.open(path, mode)  # noqa: SIM115
    elif compression == Compression.BZIP2:
        fh = bz2.open(path, mode)  # noqa: SIM115
    elif compression == Compression.LZ4:
        if not HAS_LZ4:
            raise ValueError("Python module lz4 is not available")
        fh = lz4.frame.open(path, mode)
    elif compression == Compression.ZSTD:
        if not HAS_ZSTD:
            raise ValueError("Python module backport.zstd is not available")
        fh = zstd.open(path, mode)
    elif compression == Compression.NONE:
        fh = path.open(mode)
    else:
        raise ValueError(f"Unrecognised compression method: {compression}")
    return fh


class JsonLinesWriter(JsonfileWriter):
    def __init__(self, fp: TextIO, **kwargs):
        self.fp = fp
        self.packer = SortedKeysJsonRecordPacker(indent=None)
        self.packer.on_descriptor.add_handler(self.packer_on_new_descriptor)

    def _write(self, obj: Record | RecordDescriptor) -> None:
        record_json = self.packer.pack(obj)
        line = record_json + "\n"
        line = line.encode("utf-8")
        self.fp.write(line)

    def flush(self) -> None:
        if not self.fp.closed:
            self.fp.flush()

    def close(self) -> None:
        self.fp.close()


class SortedKeysJsonRecordPacker(JsonRecordPacker):
    def pack(self, obj: Record | RecordDescriptor) -> str:
        return json.dumps(
            obj,
            default=self.pack_obj,
            indent=self.indent,
            sort_keys=True,
        )


SERIALIZERS = {
    Serialization.JSONLINES: {
        "writer": JsonLinesWriter,
        "ext": "jsonl",
    },
    Serialization.MSGPACK: {
        "writer": RecordStreamWriter,
        "ext": "rec",
    },
}


def get_sink_writer(
    full_sink_path: Path,
    serialization: Serialization,
    compression: Compression | None = None,
    new_sink: bool = True,
) -> JsonfileWriter | RecordStreamWriter:
    # create parent directories if they are missing
    full_sink_path.parent.mkdir(parents=True, exist_ok=True)

    mode = "wb" if new_sink else "ab"

    fh = open_path(full_sink_path, mode=mode, compression=compression)
    writer_cls = SERIALIZERS[serialization]["writer"]
    return writer_cls(fh)


@contextmanager
def cached_sink_writers(state: DumpState) -> Iterator[Callable]:
    # Poor man's cache that cleans up when it hits the limit of `OPEN_WRITERS_LIMIT`.
    # The cache is needed to reduce file handler open/close flickering for unsorted records stream.
    writers_cache = {}

    def close_writers() -> None:
        for path, writer in writers_cache.items():
            writer.close()
            log.debug("Sink writer closed", writer=writer, path=path)
        writers_cache.clear()

    def write_element(element: RecordStreamElement) -> int:
        sink_path = get_relative_sink_path(
            element,
            state.serialization,
            compression=state.compression,
        )
        element.sink_path = sink_path
        sink = state.get_sink(sink_path)

        new_sink = False
        if sink and not sink.is_dirty:
            log.error("Can not write to a clean finished sink", sink=sink, state=state.path)
            raise ValueError("Can not write to a clean finished sink")
        if not sink:
            sink = state.create_sink(sink_path, element)
            new_sink = True

        if sink.path in writers_cache:
            writer = writers_cache[sink.path]
        else:
            writer = writers_cache[sink.path] = get_sink_writer(
                state.get_full_sink_path(sink),
                state.serialization,
                compression=state.compression,
                new_sink=new_sink,
            )
            log.debug("Sink writer opened", writer=writer, path=sink.path)

        writer.write(element.record)
        fh_position = writer.fp.tell()

        state.update(element, fh_position)

        log.debug(
            "Record written to sink",
            sink_path=sink.path,
            record=element.record,
            writer=writer.__class__.__name__,
            position_bytes=fh_position,
            new_sink=new_sink,
        )

        if len(writers_cache) >= OPEN_WRITERS_LIMIT:
            close_writers()

        return fh_position

    try:
        yield write_element
    finally:
        close_writers()


def get_current_utc_time() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc)


def parse_datetime_iso(datetime_str: str) -> datetime.datetime:
    return datetime.datetime.fromisoformat(datetime_str)


def execute_pipeline(
    state: DumpState,
    targets: Iterator[Target],
    dry_run: bool,
    arguments: list[str],
    limit: int | None = None,
) -> None:
    """Run the record generation, processing and sinking pipeline."""

    target_func_pairs_stream = produce_target_func_pairs(targets, state)
    record_stream = itertools.islice(execute_functions(target_func_pairs_stream, dry_run, arguments), limit)
    record_stream = sink_records(record_stream, state)
    record_stream = log_progress(record_stream)
    record_stream = persist_processing_state(record_stream, state)

    # exhaust the generator, executing all pipeline steps
    deque(record_stream, maxlen=0)

    log.info("Pipeline has finished")


def parse_arguments() -> tuple[argparse.Namespace, list[str]]:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="dissect.target",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
        add_help=False,
    )
    parser.add_argument("targets", metavar="TARGET", nargs="*", help="targets to load")

    configure_plugin_arguments(parser)

    parser.add_argument(
        "-c",
        "--compression",
        choices=[c.value for c in Compression if c is not Compression.NONE],
        type=Compression,
        help="compression method",
        default=Compression.NONE,
    )
    parser.add_argument(
        "--restart",
        action="store_true",
        help="restart the session and overwrite the state file if it exists",
    )
    parser.add_argument(
        "-s",
        "--serialization",
        choices=[s.value for s in Serialization],
        default=Serialization.JSONLINES,
        type=Serialization,
        help="serialization method",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        default=".",
        required=True,
        help="output directory",
    )
    parser.add_argument("--limit", type=int, help="limit number of records produced")

    configure_generic_arguments(parser)

    args, rest = parser.parse_known_args()
    process_generic_arguments(parser, args)

    if not args.function and ("-h" in rest or "--help" in rest):
        parser.print_help()
        parser.exit(0)

    process_plugin_arguments(parser, args, rest)

    return args, rest


def main() -> None:
    args, rest = parse_arguments()

    try:
        state = configure_state(args)
        if state is None:
            # Error was already shown above, stopping execution
            return
        targets = open_targets(args)
        execute_pipeline(
            state=state,
            targets=targets,
            arguments=rest,
            dry_run=args.dry_run,
            limit=args.limit,
        )
    except Exception:
        log.exception("Exception while running the pipeline")
        sys.exit(1)


if __name__ == "__main__":
    main()



from __future__ import annotations

import re
from datetime import datetime, timezone, tzinfo
from enum import Enum, IntEnum
from typing import TYPE_CHECKING, BinaryIO, TypeVar

from dissect.util.ts import from_unix

from dissect.target.helpers import fsutil
from dissect.target.helpers.logging import get_logger

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path


log = get_logger(__name__)


def findall(buf: bytes, needle: bytes) -> Iterator[int]:
    offset = 0
    while True:
        offset = buf.find(needle, offset)
        if offset == -1:
            break

        yield offset
        offset += 1


T = TypeVar("T")


def to_list(value: T | list[T] | None) -> list[T]:
    """Convert a single value or a list of values to a list. A value of ``None`` is converted to an empty list.

    Args:
        value: The value to convert.

    Returns:
        A list of values.
    """
    if value is None:
        return []
    if not isinstance(value, list):
        return [value]

    return value


class StrEnum(str, Enum):
    """Sortable and serializible string-based enum."""


class IntEnumMissing(IntEnum):
    """Integer-based enum that allows for values other than defined members."""

    @classmethod
    def _missing_(cls, value: int) -> IntEnum:
        member = int.__new__(cls, value)
        member._name_ = str(value)
        member._value_ = value
        return member


def parse_options_string(options: str) -> dict[str, str | bool]:
    result = {}
    for opt in options.split(","):
        if "=" in opt:
            key, _, value = opt.partition("=")
            result[key] = value
        else:
            result[opt] = True
    return result


SLUG_RE = re.compile(r"[/\\ ]")


def slugify(name: str) -> str:
    """Return name with all slashes '/', backslashes '\\' and spaces ' ' replaced by underscores '_'.

    This is useful to turn a name into something that can be used as filename.
    """
    return SLUG_RE.sub("_", name)


def readinto(buffer: bytearray, fh: BinaryIO) -> int:
    """A readinto implementation that uses ``read()``.

    Reads the length of the buffer from ``fh``, and fills the buffer with said data.

    Args:
        buffer: The buffer we read the data into.
        fh: The file-like object we use for reading.

    Returns:
        the size in bytes that was read.
    """
    data = fh.read(len(buffer))
    size = len(data)
    buffer[:size] = data
    return size


STRIP_RE = re.compile(r"^[\s\x00]*|[\s\x00]*$")


def year_rollover_helper(
    path: Path, re_ts: str | re.Pattern, ts_format: str, tzinfo: tzinfo = timezone.utc
) -> Iterator[tuple[datetime, str]]:
    """Helper function for determining the correct timestamps for log files without year notation.

    Supports compressed files by using :func:`open_decompress`.

    Args:
        path: A path to the log file to parse.
        re_ts: Regex pattern for extracting the timestamp from each line.
        ts_format: Time format specification for parsing the timestamp.
        tzinfo: The timezone to use when parsing timestamps.

    Returns:
        An iterator of tuples of the parsed timestamp and the lines of the file in reverse.
    """
    # Convert the mtime to the local timezone so that we get the correct year
    mtime = path.stat().st_mtime
    current_year = from_unix(mtime).astimezone(tzinfo).year
    last_seen_month = None

    with fsutil.open_decompress(path, "rt") as fh:
        warned = False
        for line in fsutil.reverse_readlines(fh):
            line = STRIP_RE.sub(r"", line)
            if not line:
                continue

            timestamp = re.search(re_ts, line)

            if not timestamp:
                if not warned:
                    log.warning("No timestamp found in one of the lines in %s!", path)
                    warned = True
                log.debug("Skipping line: %s", line)
                continue

            # We have to append the current_year to strptime instead of adding it using replace later.
            # This prevents DeprecationWarnings on cpython >= 3.13 and Exceptions on cpython >= 3.15.
            # See https://github.com/python/cpython/issues/70647 and https://github.com/python/cpython/pull/117107.
            # Use 1904 instead of 1900 to include leap days (29 Feb).
            try:
                compare_ts = datetime.strptime(f"{timestamp.group(0)};1904", f"{ts_format};%Y").replace(tzinfo=tzinfo)
            except ValueError as e:
                log.warning("Unable to create comparison timestamp for %r in line %r: %s", timestamp.group(0), line, e)
                log.debug("", exc_info=e)
                continue

            if last_seen_month and compare_ts.month > last_seen_month:
                current_year -= 1
            last_seen_month = compare_ts.month

            try:
                relative_ts = datetime.strptime(f"{timestamp.group(0)};{current_year}", f"{ts_format};%Y").replace(
                    tzinfo=tzinfo
                )
            except ValueError as e:
                log.warning(
                    "Timestamp '%s;%s' does not match format '%s;%%Y', skipping line %r: %s",
                    timestamp.group(0),
                    current_year,
                    ts_format,
                    line,
                    e,
                )
                log.debug("", exc_info=e)
                continue

            yield relative_ts, line



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix._os import UnixPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_unix_passwd_file(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    passwd_file = absolute_path("_data/plugins/os/unix/_os/passwd")
    fs_unix.map_file("/etc/passwd", passwd_file)
    target_unix_users.add_plugin(UnixPlugin)

    results = list(target_unix_users.users())
    assert len(results) == 5
    assert results[0].source == "/etc/passwd"
    assert results[0].name == "root"
    assert results[0].passwd == "x"
    assert results[0].uid == 0
    assert results[0].gid == 0
    assert results[0].home == "/root"
    assert results[0].shell == "/bin/bash"


def test_unix_passwd_syslog(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    syslog_file = absolute_path("_data/plugins/os/unix/_os/passwd-syslog")
    fs_unix.map_file("/var/log/syslog", syslog_file)
    fs_unix.map_file_fh("/etc/passwd", BytesIO())
    target_unix_users.add_plugin(UnixPlugin)

    results = list(target_unix_users.users(sessions=True))
    assert len(results) == 3

    # first user session of john.doe
    assert results[0].source == "/var/log/syslog"
    assert results[0].name == "john.doe"
    assert results[0].home == "/home/local/john.doe"
    assert results[0].shell == "/bin/bash"

    # second user session of john.doe with different session vars
    assert results[1].source == "/var/log/syslog"
    assert results[1].name == "john.doe"
    assert results[1].home == "/root"
    assert results[1].shell == "/bin/sh"

    # first user session of jane.doe
    assert results[2].source == "/var/log/syslog"
    assert results[2].name == "jane.doe"
    assert results[2].home == "/home/local/jane.doe"
    assert results[2].shell == "/bin/zsh"


def test_unix_passwd_backup_file(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if backup passwd files are read and only unique entries are returned."""
    passwd_file = absolute_path("_data/plugins/os/unix/_os/passwd")
    fs_unix.map_file("/etc/passwd", passwd_file)
    backup_file = passwd_file.read_bytes() + b"deleted-user:x:1001:1001:deleted-user:/home/deleted:/bin/sh"
    fs_unix.map_file_fh("/etc/passwd-", BytesIO(backup_file))
    target_unix.add_plugin(UnixPlugin)

    results = list(target_unix.users())
    assert len(results) == 5 + 1



from __future__ import annotations

import hashlib
from typing import TYPE_CHECKING

import pytest

from dissect.target.helpers import keychain
from dissect.target.loader import open as loader_open
from dissect.target.loaders.ab import AndroidBackupLoader
from dissect.target.loaders.ab import main as ab_main
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target]) -> None:
    """Test that we correctly use ``AndroidBackupLoader`` when opening a ``Target``."""
    path = absolute_path("_data/loaders/ab/test.ab")
    keychain.register_wildcard_value("password")

    target = opener(path)
    assert isinstance(target._loader, AndroidBackupLoader)
    assert target.path == path


def test_loader() -> None:
    """Test the Android Backup loader."""
    path = absolute_path("_data/loaders/ab/test.ab")
    keychain.register_wildcard_value("password")

    loader = loader_open(path)
    assert isinstance(loader, AndroidBackupLoader)

    t = Target()
    loader.map(t)
    assert len(t.filesystems) == 1

    assert list(map(str, t.fs.path("/").rglob("*"))) == [
        "/data",
        "/data/data",
        "/data/data/org.fedorahosted.freeotp",
        "/data/data/org.fedorahosted.freeotp/shared_preferences",
        "/data/data/org.fedorahosted.freeotp/shared_preferences/tokenBackup.xml",
    ]

    buf = t.fs.path("/data/data/org.fedorahosted.freeotp/shared_preferences/tokenBackup.xml").read_bytes()
    assert hashlib.sha1(buf).hexdigest() == "7177d340414d5ca2a835603c347e64e3d2625e47"


def test_unwrapper(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    """Test the Android Backup unwrapper script."""
    path = absolute_path("_data/loaders/ab/test.ab")

    with monkeypatch.context() as m:
        m.setattr("sys.argv", ["python", str(path), "-p", "password", "-o", str(tmp_path)])
        ab_main()

        assert [entry.name for entry in tmp_path.iterdir()] == ["test.plain.ab"]

        m.setattr("sys.argv", ["python", str(path), "-p", "password", "-o", str(tmp_path), "-t"])
        ab_main()

        assert sorted([entry.name for entry in tmp_path.iterdir()]) == ["test.plain.ab", "test.tar"]



from __future__ import annotations

import textwrap
from datetime import datetime, timezone
from io import BytesIO
from typing import TYPE_CHECKING
from unittest.mock import patch

from flow.record.fieldtypes import posix_path

from dissect.target.plugins.os.unix.bsd.citrix._os import CitrixPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_citrix_os(target_citrix: Target, fs_bsd: VirtualFilesystem) -> None:
    example_etc_passwd = """
    # $FreeBSD: releng/11.4/etc/master.passwd 359448 2020-03-30 17:11:21Z brooks $
    #
    root:*:0:0:Charlie &:/root:/usr/bin/bash
    bind:*:53:53:Bind Sandbox:/:/usr/sbin/nologin
    nobody:*:65534:65534:Unprivileged user:/nonexistent:/usr/sbin/nologin
    """

    fs_bsd.map_file_fh("/root/.cli_history", BytesIO(b'echo "hello world"'))
    fs_bsd.map_file_fh("/var/nstmp/robin/.cli_history", BytesIO(b'echo "hello world"'))
    fs_bsd.map_file_fh("/var/nstmp/alfred/.cli_history", BytesIO(b'echo "bye world"'))
    fs_bsd.map_file_fh("/etc/passwd", BytesIO(textwrap.dedent(example_etc_passwd).encode()))

    target_citrix.add_plugin(CitrixPlugin)

    assert target_citrix.os == "citrix-netscaler"
    hostname = target_citrix.hostname
    version = target_citrix.version
    users = sorted(target_citrix.users(), key=lambda user: (user.name, str(user.home) if user.home else ""))
    ips = target_citrix.ips
    ips.sort()

    assert hostname == "mynetscaler"
    assert version == "NetScaler 13.1 build 30 (ns-13.1-30.52)"

    assert ips == ["10.0.0.68", "10.0.0.69"]

    assert target_citrix.timezone == "Europe/Amsterdam"

    entry = target_citrix.filesystems[2].get(".version")
    stat_result = entry.stat()
    stat_result.st_birthtime = 1234567890
    with patch.object(entry, "stat") as mock_stat:
        mock_stat.return_value = stat_result
        assert target_citrix.install_date == datetime(2009, 2, 13, 23, 31, 30, tzinfo=timezone.utc)

    target_citrix.filesystems[1].map_file_fh(
        "netscaler/logon/themes/SomeThemeName/SomeFile.json", BytesIO(b'{"locale": "en"}')
    )
    assert target_citrix.language == ["en_US"]

    assert len(users) == 8

    assert users[0].name == "alfred"  # Only listed in /var/nstmp
    assert users[0].home == posix_path("/var/nstmp/alfred")

    assert users[1].name == "batman"  # Only listed in config
    assert users[1].home is None

    assert users[2].name == "bind"  # User entry from /etc/passwd, home overwritten from '/' to None
    assert users[2].home is None

    assert users[3].name == "jasontodd"  # Only listed in config backup
    assert users[3].home is None

    assert users[4].name == "nobody"  # User entry for the nobody user from /etc/passwd
    assert users[4].home == posix_path("/nonexistent")

    assert users[5].name == "robin"  # Listed in config and /var/nstmp
    assert users[5].home == posix_path("/var/nstmp/robin")

    assert users[6].name == "root"  # User entry for /root, from the config
    assert users[6].home == posix_path("/root")
    assert users[6].shell is None

    assert users[7].name == "root"  # User entry for /root, from /etc/passwd
    assert users[7].home == posix_path("/root")
    assert users[7].shell == "/usr/bin/bash"



from __future__ import annotations

import io
import json
import re
import sys
from collections import deque
from collections.abc import Callable, ItemsView, Iterable, Iterator, KeysView
from configparser import ConfigParser, MissingSectionHeaderError
from contextlib import nullcontext
from dataclasses import dataclass
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    TextIO,
)

from defusedxml import ElementTree

from dissect.target.exceptions import ConfigurationParsingError, FileNotFoundError
from dissect.target.helpers.logging import get_logger
from dissect.target.helpers.utils import to_list

if TYPE_CHECKING:
    from pathlib import Path
    from types import TracebackType

    from typing_extensions import Self

try:
    from ruamel.yaml import YAML

    HAS_YAML = True
except ImportError:
    HAS_YAML = False

try:
    if sys.version_info < (3, 11):
        import tomli as toml
    else:
        # tomllib is included since python 3.11
        import tomllib as toml  # novermin

    HAS_TOML = True
except ImportError:
    HAS_TOML = False


log = get_logger(__name__)


def _update_dictionary(current: dict[str, Any], key: str, value: Any) -> None:
    if (prev_value := current.get(key)) is not None:  #  "" is a value
        if isinstance(prev_value, dict) and isinstance(value, dict):
            prev_value.update(value)
            return

        if isinstance(prev_value, str):
            prev_value = [prev_value]

        if isinstance(prev_value, list):
            # We want to append ``value`` to prev_value
            prev_value.append(value)

    current[key] = prev_value or value


class PeekableIterator:
    # https://more-itertools.readthedocs.io/en/stable/_modules/more_itertools/more.html#peekable

    def __init__(self, iterable: Iterable[str]):
        self._iterator = iter(iterable)
        self._cache = deque()

    def __iter__(self) -> PeekableIterator:
        return self

    def __next__(self) -> str:
        if self._cache:
            return self._cache.popleft()

        return next(self._iterator)

    def peek(self) -> str | None:
        if not self._cache:
            try:
                self._cache.append(next(self._iterator))
            except StopIteration:
                return None
        return self._cache[0]


class ConfigurationParser:
    """A configuration parser where you can configure certain aspects of the parsing mechanism.

    Args:
        collapse: A ``bool`` or an ``Iterator``:
          If ``True``: it will collapse all the resulting dictionary values.
          If an ``Iterable`` it will collapse on the keys defined in ``collapse``.
        collapse_inverse: Inverses the collapsing mechanism. Collapse on everything that is not inside ``collapse``.
        separator: Contains what values it should look for as a separator.
        comment_prefixes: Contains what constitutes as a comment.
    """

    def __init__(
        self,
        collapse: bool | Iterable[str] = False,
        collapse_inverse: bool = False,
        separator: tuple[str] = ("=",),
        comment_prefixes: tuple[str] = (";", "#"),
    ) -> None:
        self.collapse_all = collapse is True
        self.collapse = set(collapse) if isinstance(collapse, Iterable) else set()
        self._collapse_check = self._key_not_in_collapse if collapse_inverse else self._key_in_collapse

        self.separator = separator
        self.comment_prefixes = comment_prefixes
        self.parsed_data = {}

    def __getitem__(self, item: Any) -> dict | str:
        return self.parsed_data[item]

    def __contains__(self, item: str) -> bool:
        return item in self.parsed_data

    def _collapse_dict(self, dictionary: dict, collapse: bool = False) -> dict[str, dict]:
        new_dictionary = {}

        if isinstance(dictionary, list) and collapse:
            return dictionary[-1]

        if not hasattr(dictionary, "items"):
            return dictionary

        for key, value in dictionary.items():
            value = self._collapse_dict(value, self.collapse_all or self._collapse_check(key))
            new_dictionary.update({key: value})

        return new_dictionary

    def _key_in_collapse(self, key: str) -> bool:
        return key in self.collapse

    def _key_not_in_collapse(self, key: str) -> bool:
        return key not in self.collapse

    def parse_file(self, fh: TextIO) -> None:
        """Parse the contents of ``fh`` into key/value pairs.

        This function should **set** :attr:`parsed_data` as a side_effect.

        Args:
            fh: The text to parse.
        """
        raise NotImplementedError

    def get(self, item: str, default: Any | None = None) -> Any:
        return self.parsed_data.get(item, default)

    def read_file(self, fh: TextIO | io.BytesIO) -> None:
        """Parse a configuration file.

        Raises:
            ConfigurationParsingError: If any exception occurs during during the parsing process.
        """

        try:
            self.parse_file(fh)
        except Exception as e:
            raise ConfigurationParsingError(e.args) from e

        if self.collapse_all or self.collapse:
            self.parsed_data = self._collapse_dict(self.parsed_data)

        if not isinstance(self.parsed_data, dict):
            self.parsed_data = self._collapse_dict(self.parsed_data, False)

    def merge(self, other: ConfigurationParser) -> ConfigurationParser:
        """Merge the contents of another parser into this one.
        On conflict, the values of the other parser will be used.

        Args:
            other: The other parser to merge.

        Returns:
            The merged parser.
        """

        self._merge(self.parsed_data, other.parsed_data)
        return self

    def _merge(self, dict1: dict, dict2: dict) -> dict:
        for key, value2 in dict2.items():
            value1 = dict1.get(key)
            if value1 is None:
                # Result does not have key yet, add it
                dict1[key] = value2
                continue

            collapse = self.collapse_all or self._collapse_check(key)

            if collapse:
                if isinstance(value1, dict) and isinstance(value2, dict):
                    self._merge(value1, value2)  # There should only be one, merge both
                else:
                    dict1[key] = value2
            else:
                combined = to_list(value1) + to_list(value2)
                # An empty string clears the list of values for the current key
                # Possibly turn this into an option if other parsers require different behavior
                combined = combined[combined.index("") + 1 :] if "" in combined else combined
                dict1[key] = combined

        return dict1

    def keys(self) -> KeysView:
        return self.parsed_data.keys()

    def items(self) -> ItemsView:
        return self.parsed_data.items()

    def as_dict(self) -> dict:
        return self.parsed_data


class Default(ConfigurationParser):
    """Parse a configuration file specified by ``separator`` and ``comment_prefixes``.

    This parser splits only on the first ``separator`` it finds:

    .. code-block::

        key<separator>value     -> {"key": "value"}

        key<separator>value\n
          continuation
                                -> {"key": "value continuation"}

        # Unless we collapse values, we add them to a list to not overwrite any values.
        key<separator>value1
        key<separator>value2
                                -> {key: [value1, value2]}

        <empty_space><comment>  -> skip
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.SEPARATOR = re.compile(rf"\s*[{''.join(self.separator)}]\s*")
        self.COMMENTS = re.compile(rf"\s*[{''.join(self.comment_prefixes)}]")
        self.skip_lines = (*self.comment_prefixes, "\n")

    def line_reader(self, fh: TextIO, strip_comments: bool = True) -> Iterator[str]:
        for line in fh:
            if line.strip().startswith(self.skip_lines) or not line.strip():
                continue

            if strip_comments:
                line, *_ = self.COMMENTS.split(line, 1)

            yield line

    def parse_file(self, fh: TextIO) -> None:
        information_dict = {}

        prev_key = None
        for line in self.line_reader(fh):
            if line.startswith((" ", "\t")):
                # This part was indented so it is assumed to be a continuation of the previous key
                prev_value = information_dict.get(prev_key)
                if not isinstance(prev_value, str):
                    # If the previous value is not a string, some parsing already occured which would imply
                    # that this line is not a continuation of the previous value.
                    # Meaning we are in an unknown state and should not continue parsing.

                    raise ConfigurationParsingError(
                        "Unexpected state occurred when parsing using the `Default` parser."
                    )
                information_dict[prev_key] = f"{prev_value} {line.strip()}"
                continue

            prev_key, *value = self.SEPARATOR.split(line, 1)
            value = value[0].strip() if value else ""

            _update_dictionary(information_dict, prev_key, value)

        self.parsed_data = information_dict


class CSVish(Default):
    """Parses CSV-ish config files (does not confirm to CSV standard!)."""

    def __init__(self, *args, fields: tuple[str, ...], **kwargs):
        self.fields = fields
        self.num_fields = len(self.fields)
        self.maxsplit = self.num_fields - 1
        super().__init__(*args, **kwargs)

    def parse_file(self, fh: TextIO) -> None:
        information_dict = {}

        for i, raw_line in enumerate(self.line_reader(fh, strip_comments=True)):
            line = raw_line.strip()
            columns = re.split(self.SEPARATOR, line, maxsplit=self.maxsplit)

            # keep unparsed lines separate (often env vars)
            data = {"line": line} if len(columns) < self.num_fields else dict(zip(self.fields, columns, strict=False))

            information_dict[str(i)] = data

        self.parsed_data = information_dict


class Ini(ConfigurationParser):
    """Parses an ini file according using the built-in Python ``ConfigParser``."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.parsed_data = ConfigParser(
            strict=False,
            delimiters=self.separator,
            comment_prefixes=self.comment_prefixes,
            allow_no_value=True,
            interpolation=None,
        )
        self.parsed_data.optionxform = str

    def parse_file(self, fh: TextIO) -> None:
        offset = fh.tell()
        try:
            return self.parsed_data.read_file(fh)
        except MissingSectionHeaderError:
            pass

        fh.seek(offset)
        open_file = io.StringIO("[DEFAULT]\n" + fh.read())
        return self.parsed_data.read_file(open_file)


class Txt(ConfigurationParser):
    """Read the file into ``content``, and show the bumber of bytes read."""

    def parse_file(self, fh: TextIO) -> None:
        # Cast the size to a string, to print it out later.
        self.parsed_data = {"content": fh.read(), "size": str(fh.tell())}


class Bin(ConfigurationParser):
    """Read the file into ``binary`` and show the number of bytes read."""

    def parse_file(self, fh: io.BytesIO) -> None:
        self.parsed_data = {"binary": fh.read(), "size": str(fh.tell())}


class Xml(ConfigurationParser):
    """Parses an XML file. Ignores any constructor parameters passed from ``ConfigurationParser``."""

    def _tree(self, tree: ElementTree, root: bool = False) -> dict:
        """Very simple but robust xml -> dict implementation, see comments."""
        nodes = {}
        result = {}
        counter = {}

        # each node is a folder (so the structure is always the same! [1])
        for node in tree.findall("*"):
            # if a node contains multiple nodes with the same name, number them
            if node.tag in counter:
                counter[node.tag] += 1
                nodes[f"{node.tag}-{counter[node.tag]}"] = self._tree(node)
            else:
                counter[node.tag] = 1
                nodes[node.tag] = self._tree(node)

        # all attribs go in the attribute folder
        # (i.e. stable, does not change depending on xml structure! [2]
        # Also, this way we "know" they have been attributes, i.e. we don't lose information! [3]
        if tree.attrib:
            result["attributes"] = tree.attrib

        # all subnodes go in the nodes folder
        if nodes:
            result["nodes"] = nodes

        # content goes into the text folder
        # we don't use special prefixes ($) because XML docs may use them anyway (even though they are forbidden)
        if tree.text and (text := tree.text.strip(" \n\r")):
            result["text"] = text

        # if you need to store meta-data, you can extend add more entries here... CDATA, Comments, errors
        return {tree.tag: result} if root else result

    def _fix(self, content: str, position: tuple[int, int]) -> str:
        """Quick heuristic fix. If there is an invalid token, just remove it."""
        lineno, offset = position
        lines = content.split("\n")

        line = lines[lineno - 1]
        line = line[: offset - 1] + "" + line[offset + 1 :]

        lines[lineno - 1] = line

        return "\n".join(lines)

    def parse_file(self, fh: TextIO) -> None:
        content = fh.read()
        document = content
        errors = 0
        limit = 20
        tree = {}

        while not tree and errors < limit:
            try:
                tree = self._tree(ElementTree.fromstring(document), root=True)
                break
            except ElementTree.ParseError as e:
                errors += 1
                document = self._fix(document, e.position)

        if not tree:
            # Error limit reached. Thus we consider the document not parseable.
            raise ConfigurationParsingError(f"Could not parse XML file: {fh.name} after {errors} attempts.")

        self.parsed_data = tree


class ListUnwrapper:
    """Provides utility functions to unwrap dictionary objects out of lists."""

    @staticmethod
    def unwrap(data: dict | list) -> dict | list:
        """Transforms a list with dictionaries to a dictionary.

        The order of the list is preserved. If no dictionary is found, the list remains untouched:

        .. code-block::

            ["value1", "value2"]    -> ["value1", "value2"]

            {"data": "value"}       -> {"data": "value"}

            [{"data": "value"}]     -> {
                                           "list_item0": {
                                                "data": "value"
                                           }
                                       }
        """
        orig = ListUnwrapper._unwrap_dict_list(data)
        return ListUnwrapper._unwrap_dict(orig)

    @staticmethod
    def _unwrap_dict(data: dict | list) -> dict | list:
        """Looks for dictionaries and unwraps its values."""

        if not isinstance(data, dict):
            return data

        root = {}
        for key, value in data.items():
            _value = ListUnwrapper._unwrap_dict_list(value)
            if isinstance(_value, dict):
                _value = ListUnwrapper._unwrap_dict(_value)
            root[key] = _value

        return root

    @staticmethod
    def _unwrap_dict_list(data: dict | list) -> dict | list:
        """Unwraps a list containing dictionaries."""
        if not isinstance(data, list) or not any(isinstance(obj, dict) for obj in data):
            return data

        return_value = {}
        for idx, elem in enumerate(data):
            return_value[f"list_item{idx}"] = elem

        return return_value


class Json(ConfigurationParser):
    """Parses a JSON file."""

    def parse_file(self, fh: TextIO) -> None:
        parsed_data = json.load(fh)
        self.parsed_data = ListUnwrapper.unwrap(parsed_data)


class Yaml(ConfigurationParser):
    """Parses a Yaml file."""

    def parse_file(self, fh: TextIO) -> None:
        if HAS_YAML:
            parsed_data = YAML(typ="safe").load(fh)
            self.parsed_data = ListUnwrapper.unwrap(parsed_data)
        else:
            raise ConfigurationParsingError("Failed to parse file, please install ruamel.yaml.")


class Toml(ConfigurationParser):
    """Parses a Toml file."""

    def parse_file(self, fh: TextIO) -> None:
        if HAS_TOML:
            self.parsed_data = toml.loads(fh.read())
        else:
            raise ConfigurationParsingError("Failed to parse file, please install tomli.")


class Env(ConfigurationParser):
    """Parses ``.env`` file contents according to Docker and bash specification.

    Does not apply interpolation of substituted values, e.g. ``foo=${bar}`` and does not attempt to parse list or dict
    strings. Does not support dynamic env files, e.g. ``foo=`bar```. Also does not support multi-line key/value
    assignments (yet).

    References:
        - https://docs.docker.com/compose/environment-variables/variable-interpolation/#env-file-syntax
        - https://github.com/theskumar/python-dotenv/blob/main/src/dotenv/parser.py
    """

    RE_KV = re.compile(r"^(?P<key>.+?)=(?P<value>(\".+?\")|(\'.+?\')|(.*?))?(?P<comment> \#.+?)?$")

    def __init__(self, comments: bool = False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.comments = comments
        self.parsed_data: dict | tuple[dict, str | None] = {}

    def parse_file(self, fh: TextIO) -> None:
        for line in fh:
            # Blank lines are ignored.
            # Lines beginning with ``#`` are processed as comments and ignored.
            if not line or line[0] == "#" or "=" not in line:
                continue

            # Each line represents a key-value pair. Values can optionally be quoted.
            # Inline comments for unquoted values must be preceded with a space.
            # Value may be empty.
            match = self.RE_KV.match(line)

            # Line could be invalid
            if not match:
                log.warning("Could not parse line in %s: '%s'", fh, line)
                continue

            key = match.groupdict()["key"]
            value = match.groupdict().get("value") or ""
            value = value.strip()
            comment = match.groupdict().get("comment")
            comment = comment.replace(" # ", "", 1) if comment else None

            # Surrounding whitespace characters are removed, unless quoted.
            if value and ((value[0] == '"' and value[-1] == '"') or (value[0] == "'" and value[-1] == "'")):
                is_quoted = True
                value = value.strip("\"'")
            else:
                is_quoted = False
                value = value.strip()

            # Unquoted values may start with a quote if they are properly escaped.
            if not is_quoted and value[:2] in ["\\'", '\\"']:
                value = value[1:]

            # Interpret boolean values
            if value.lower() in ["1", "true"]:
                value = True
            elif value.lower() in ["0", "false"]:
                value = False

            # Interpret integer values
            if isinstance(value, str) and re.match(r"^[0-9]{1,}$", value):
                value = int(value)

            if key.strip() in self.parsed_data:
                log.warning("Duplicate environment key '%s' in file %s", key.strip(), fh)

            self.parsed_data[key.strip()] = (value, comment) if self.comments else value


class ScopeManager:
    """A (context)manager for dictionary scoping.

    This class provides utility functions to keep track of scopes inside a dictionary.

    Attributes:
        _parents: A dictionary accounting what child belongs to which parent dictionary.
        _root: The initial dictionary.
        _current: The current dictionary.
        _previous: The node before the current (changed) node.
    """

    def __init__(self):
        self._parents = {}
        self._root = {}
        self._current = self._root
        self._previous = None

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        type: type[BaseException] | None,
        value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self.clean()

    def _set_prev(self, keep_prev: bool) -> None:
        """Set :attr:`_previous` before :attr:`_current` changes."""
        if not keep_prev:
            self._previous = self._current

    def push(self, name: str, keep_prev: bool = False) -> Literal[True]:
        """Push a new key to the :attr:`_current` dictionary and return that we did."""
        child = self._current.get(name)
        new_child = {}
        parent = self._current
        if isinstance(child, dict):
            # A second scope with the same name, turn the existing one into a list and append a new one
            parent[name] = [child, new_child]
            child = parent[name]
        elif isinstance(child, list):
            # Multiple scopes with the same name, append a new one to the list
            child.append(new_child)
        elif isinstance(child, str):
            # Child is not a scope but a scalar value, do nothing
            pass
        else:
            # Create a new scope
            parent[name] = new_child

        self._parents[id(new_child)] = parent
        self._set_prev(keep_prev)
        self._current = new_child
        return True

    def pop(self, keep_prev: bool = False) -> bool:
        """Pop :attr:`_current` and return whether we changed the :attr:`_parents` dictionary."""
        if new_current := self._parents.pop(id(self._current), None):
            self._set_prev(keep_prev)
            self._current = new_current
            return True
        return False

    def update(self, key: str, value: str) -> None:
        """Update the :attr:`_current` dictionary with ``key`` and ``value``."""
        _update_dictionary(self._current, key, value)

    def update_prev(self, key: str, value: str) -> None:
        """Update the :attr:`_previous` dictionary with ``key`` and ``value``."""
        _update_dictionary(self._previous, key, value)

    def is_root(self) -> bool:
        """Utility function to check whether the current dictionary is a root dictionary."""
        return id(self._current) == id(self._root)

    def clean(self) -> None:
        """Clean up the internal state.
        This is called automatically when :class:`ScopeManager` is used as a contextmanager.
        """
        self._parents = {}
        self._root = {}
        self._current = self._root
        self._previous = None


class Indentation(Default):
    """This parser is used for files that use a single level of indentation to specify a different scope.

    Examples of these files are the ``sshd_config`` file.
    Where "Match" statements use a single layer of indentation to specify a scope for the key value pairs.

    The parser parses this as the following:

    .. code-block::

      key value
        key2 value2
                       -> {"key value": {"key2": "value2"}}
    """

    def _parse_line(self, line: str) -> tuple[str, str]:
        key, *value = self.SEPARATOR.split(line.strip(), 1)
        value = value[0].strip() if value else ""
        return key, value

    def _change_scope(
        self,
        manager: ScopeManager,
        line: str,
        key: str,
        next_line: str | None = None,
    ) -> bool:
        """A function to check whether to create a new scope, or go back to a previous one.

        Args:
            manager: A :class:`ScopeManager` that contains the logic to ``push`` and ``pop`` scopes. And keeps state.
            line: The line to be parsed.
            key: The key that should be updated during a :method:`ScopeManager.push`.
            next_line: The next line to be parsed.

        Returns:
            Whether the scope changed or not.
        """
        empty_space = (" ", "\t")
        changed = False

        if next_line is None:
            return False

        if not line.startswith(empty_space):
            changed = manager.pop()

        if not line.startswith(empty_space) and next_line.startswith(empty_space):
            return manager.push(key)
        return changed

    def parse_file(self, fh: TextIO) -> None:
        iterator = PeekableIterator(self.line_reader(fh))
        prev_key = None

        with ScopeManager() as manager:
            for line in iterator:
                key, value = self._parse_line(line)
                changed = self._change_scope(
                    manager=manager,
                    line=line,
                    key=line.strip(),
                    next_line=iterator.peek(),
                )

                if changed:
                    prev_key = line.strip()
                    continue

                if not value:
                    key, value = prev_key, key
                    manager.pop()

                manager.update(key, value)

            self.parsed_data = manager._root


class SystemD(Indentation):
    """A :class:`ConfigurationParser` that specifically parses systemd configuration files.

    Examples:

        .. code-block::

            >>> systemd_data = textwrap.dedent(
                    '''
                    [Section1]
                    Key=Value
                    [Section2]
                    Key2=Value 2\\
                        Value 2 continued
                    '''
                )
            >>> parser = SystemD(io.StringIO(systemd_data))
            >>> parser.parser_items
            {
                "Section1": {
                    "Key": "Value
                },
                "Section2": {
                    "Key2": "Value2 Value 2 continued
                }
            }
    """

    def _change_scope(
        self,
        manager: ScopeManager,
        line: str,
        key: str,
        next_line: str | None = None,
    ) -> bool:
        scope_char = ("[", "]")
        changed = False
        if line.lstrip().startswith(scope_char):
            if not manager.is_root():
                changed = manager.pop()
            stripped_characters = "".join(scope_char)
            changed = manager.push(key.strip(stripped_characters), changed)

        return changed

    def parse_file(self, fh: TextIO) -> None:
        prev_values = []
        prev_key = None

        with ScopeManager() as manager:
            for line in self.line_reader(fh, strip_comments=False):
                changed = self._change_scope(
                    manager=manager,
                    line=line,
                    key=line.strip(),
                )

                if changed:
                    # Current part is a section header.
                    if prev_values:
                        # Update previous key/value... someone configured it wrong
                        prev_values, prev_key = self._update_continued_values(
                            func=manager.update_prev,
                            key=prev_key,
                            values=prev_values,
                        )
                    continue

                key, value = self._parse_line(line)

                continued_value = value or key
                if continued_value.endswith("\\"):
                    prev_key = prev_key or key
                    prev_values.append(continued_value.strip("\\ "))
                    continue

                if prev_values:
                    prev_values, prev_key = self._update_continued_values(
                        func=manager.update,
                        key=prev_key,
                        values=[*prev_values, continued_value],
                    )
                    continue

                manager.update(key, value)

            self.parsed_data = manager._root

    def _update_continued_values(
        self, func: Callable[[str, str], None], key: str, values: list[str]
    ) -> tuple[list, None]:
        value = " ".join(values)
        func(key, value)
        return [], None


class Leases(Default):
    """A :class:`ConfigurationParser` that specifically parses dhclient ``.leases`` files.

    Examples:

        .. code-block::

            >>> Leases = textwrap.dedent(
                    '''
                    lease {
                        interface "eth0"; # A comment that gets ignored
                        fixed-address 1.2.3.4;
                        option dhcp-lease-time 13337;
                        option routers 0.0.0.0;
                        option host-name "hostname";
                        renew 1 2023/12/31 13:37:00;
                        rebind 2 2023/01/01 01:00:00;
                        expire 3 2024/01/01 13:37:00;
                        # Another comment that gets ignored
                    }
                    '''
                )
            >>> parser = Leases(io.StringIO(lease))
            >>> parser.parsed_data
            {
                "lease-0": {
                    "interface": "eth0",
                    "fixed-address": "1.2.3.4"
                    "option": {
                        "dhcp-lease-time": "13337",
                        "routers": "0.0.0.0",
                        "host-name": "hostname"
                    },
                    "renew": "1 2023/12/31 13:37:00",
                    "rebind": "2 2023/01/01 01:00:00",
                    "expire": "3 2024/01/01 13:37:00",
            }
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, separator=(r"\s",))

    def _parse_line(self, line: str) -> tuple[str, str]:
        key, *value = self.SEPARATOR.split(line.strip(), 1)
        value = value[0].strip() if value else ""
        return key, value

    def parse_file(self, fh: TextIO) -> None:
        root = None

        for line in self.line_reader(fh):
            key, value = self._parse_line(line)

            if key.startswith("}"):
                continue

            if key.startswith("lease"):
                idx = len([key for key in self.parsed_data if "lease" in key])
                root = f"{key}-{idx}"
                self.parsed_data[root] = {}
                continue

            if root:
                if key == "option":
                    value = dict([value.split(maxsplit=1)])

                _update_dictionary(self.parsed_data[root], key, value)
            else:
                _update_dictionary(self.parsed_data, key, value)


@dataclass(frozen=True)
class ParserOptions:
    collapse: bool | set[str] | None = None
    collapse_inverse: bool | None = None
    separator: tuple[str] | None = None
    comment_prefixes: tuple[str] | None = None


@dataclass(frozen=True)
class ParserConfig:
    parser: type[ConfigurationParser] = Default
    collapse: bool | set[str] | None = None
    collapse_inverse: bool | None = None
    separator: tuple[str] | None = None
    comment_prefixes: tuple[str] | None = None
    fields: tuple[str, ...] | None = None

    def create_parser(self, options: ParserOptions | None = None) -> ConfigurationParser:
        kwargs = {}

        for field_name in ["collapse", "collapse_inverse", "separator", "comment_prefixes", "fields"]:
            value = getattr(options, field_name, None) or getattr(self, field_name)
            if value:
                kwargs.update({field_name: value})

        return self.parser(**kwargs)


MATCH_MAP: dict[str, ParserConfig] = {
    "*/systemd/*": ParserConfig(SystemD),
    "*/sysconfig/network-scripts/ifcfg-*": ParserConfig(Default),
    "*/sysctl.d/*.conf": ParserConfig(Default),
    "*/xml/*": ParserConfig(Xml),
    "*.bashrc": ParserConfig(Txt),
    "*/vim/vimrc*": ParserConfig(Txt),
    "*/fonts/conf.*/*.conf": ParserConfig(Xml),
}

CONFIG_MAP: dict[str, ParserConfig] = {
    "ini": ParserConfig(Ini),
    "env": ParserConfig(Env),
    "xml": ParserConfig(Xml),
    "json": ParserConfig(Json),
    "yml": ParserConfig(Yaml),
    "yaml": ParserConfig(Yaml),
    "cnf": ParserConfig(Default),
    "conf": ParserConfig(Default, separator=(r"\s",)),
    "sample": ParserConfig(Txt),
    "sh": ParserConfig(Txt),
    "key": ParserConfig(Txt),
    "crt": ParserConfig(Txt),
    "pem": ParserConfig(Txt),
    "pl": ParserConfig(Txt),  # various admin panels
    "lua": ParserConfig(Txt),  # wireshark etc.
    "txt": ParserConfig(Txt),
    "systemd": ParserConfig(SystemD),
    "template": ParserConfig(Txt),
    "toml": ParserConfig(Toml),
    "leases": ParserConfig(Leases),
    "meta_bare": ParserConfig(Default),  # Return the basic config parser
}


KNOWN_FILES: dict[str, ParserConfig] = {
    "ulogd.conf": ParserConfig(Ini),
    "sshd_config": ParserConfig(Indentation, separator=(r"\s",)),
    "sensors3.conf": ParserConfig(Indentation, separator=(r"\s",)),
    "hosts.allow": ParserConfig(Default, separator=(":",), comment_prefixes=("#",)),
    "hosts.deny": ParserConfig(Default, separator=(":",), comment_prefixes=("#",)),
    "hosts": ParserConfig(Default, separator=(r"\s",)),
    "nsswitch.conf": ParserConfig(Default, separator=(":",)),
    "lsb-release": ParserConfig(Default),
    "catalog": ParserConfig(Xml),
    "ld.so.cache": ParserConfig(Bin),
    "fstab": ParserConfig(
        CSVish,
        separator=(r"\s",),
        comment_prefixes=("#",),
        fields=("device", "mount", "type", "options", "dump", "pass"),
    ),
    "crontab": ParserConfig(
        CSVish,
        separator=(r"\s",),
        comment_prefixes=("#",),
        fields=("minute", "hour", "day", "month", "weekday", "user", "command"),
    ),
    "shadow": ParserConfig(
        CSVish,
        separator=(r"\:",),
        comment_prefixes=("#",),
        fields=(
            "username",
            "password",
            "lastchange",
            "minpassage",
            "maxpassage",
            "warning",
            "inactive",
            "expire",
            "rest",
        ),
    ),
    "passwd": ParserConfig(
        CSVish,
        separator=(r"\:",),
        comment_prefixes=("#",),
        fields=("username", "password", "uid", "gid", "gecos", "homedir", "shell"),
    ),
    "mime.types": ParserConfig(CSVish, separator=(r"\s+",), comment_prefixes=("#",), fields=("name", "extensions")),
}


def parse(path: Path, hint: str | None = None, *args, **kwargs) -> ConfigurationParser:
    """Parses the content of an ``path`` or ``entry`` to a dictionary.

    Args:
        path: The path to either a directory or file.
        hint: What kind of parser should be used.
        collapse: Whether it should collapse everything or just a certain set of keys.
        collapse_inverse: Invert the collapse function to collapse everything but the keys inside ``collapse``.
        separator: The separator that should be used for parsing.
        comment_prefixes: What is specified as a comment.

    Raises:
        FileNotFoundError: If the ``path`` is not a file.
    """

    if not path.is_file():
        raise FileNotFoundError(f"Could not parse {path} as a dictionary.")

    options = ParserOptions(*args, **kwargs)
    return parse_config(path, hint, options)


def parse_config(
    entry: Path,
    hint: str | None = None,
    options: ParserOptions | None = None,
) -> ConfigurationParser:
    parser_type = _select_parser(entry, hint)

    parser = parser_type.create_parser(options)

    cm = nullcontext if isinstance(parser, Bin) else partial(io.TextIOWrapper, encoding="utf-8")

    with entry.open("rb") as fh, cm(fh) as open_file:
        parser.read_file(open_file)

    if isinstance(parser, SystemD):
        return _parse_drop_files(entry, options, parser)

    return parser


def _parse_drop_files(path: Path, options: ParserOptions, main_parser: ConfigurationParser) -> ConfigurationParser:
    if not (drop_folder := path.with_name(path.name + ".d")).exists():
        return main_parser

    for drop_file in sorted(drop_folder.glob("*.conf")):
        if not drop_file.is_file():
            continue

        drop_file_parser = ParserConfig(SystemD).create_parser(options)
        with drop_file.open("r") as fh:
            drop_file_parser.read_file(fh)
            main_parser.merge(drop_file_parser)

    return main_parser


def _select_parser(path: Path, hint: str | None = None) -> ParserConfig:
    if hint and (parser_type := CONFIG_MAP.get(hint)):
        return parser_type

    for match, value in MATCH_MAP.items():
        if path.match(match):
            return value

    extention_parser = CONFIG_MAP.get(path.suffix.lstrip("."), ParserConfig(Txt))
    return KNOWN_FILES.get(path.name, extention_parser)



#!/usr/bin/env python
from __future__ import annotations

import argparse
import functools
import json
import logging
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any

from dissect.target.exceptions import TargetError
from dissect.target.helpers.logging import get_logger
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.tools.query import record_output
from dissect.target.tools.utils.cli import (
    catch_sigpipe,
    configure_generic_arguments,
    open_targets,
    process_generic_arguments,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from dissect.target.target import Target


InfoRecord = TargetRecordDescriptor(
    "target/info",
    [
        ("datetime", "last_activity"),
        ("datetime", "install_date"),
        ("net.ipaddress[]", "ips"),
        ("string", "os_family"),
        ("string", "os_version"),
        ("string", "architecture"),
        ("string[]", "language"),
        ("string", "timezone"),
        ("string[]", "disks"),
        ("string[]", "volumes"),
        ("string[]", "mounts"),
        ("string[]", "children"),
    ],
)


log = get_logger(__name__)
logging.lastResort = None
logging.raiseExceptions = False


@catch_sigpipe
def main() -> int:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="target-info",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
    )
    parser.add_argument("targets", metavar="TARGETS", nargs="*", help="Targets to display info from")
    parser.add_argument("--from-file", nargs="?", type=Path, help="file containing targets to load")
    parser.add_argument("-s", "--strings", action="store_true", help="print output as string")
    parser.add_argument("-r", "--record", action="store_true", help="print output as record")
    parser.add_argument("-j", "--json", action="store_true", help="output records as pretty json")
    parser.add_argument("-J", "--jsonlines", action="store_true", help="output records as one-line json")
    configure_generic_arguments(parser)

    args, _ = parser.parse_known_args()

    if not args.targets and not args.from_file:
        parser.error("too few arguments")

    if args.from_file:
        if not args.from_file.exists():
            parser.error(f"--from-file {args.from_file} does not exist")

        targets = args.from_file.read_text().splitlines()
        while targets[-1] == "":
            targets = targets[:-1]
        args.targets = targets

    process_generic_arguments(parser, args)

    try:
        for i, target in enumerate(open_targets(args)):
            try:
                target_info = get_target_info(target, args.recursive)
                if args.jsonlines:
                    print(json.dumps(target_info, default=str))
                elif args.json:
                    print(json.dumps(target_info, indent=4, default=str))
                elif args.record:
                    rs = record_output(args.strings)
                    rs.write(InfoRecord(**target_info, _target=target))
                else:
                    if i > 0:
                        print("-" * 70)
                    print_target_info(target, target_info)
            except Exception as e:  # noqa: PERF203
                target.log.error("Exception in retrieving information for target: `%s`, use `-vv` for details", target)  # noqa: TRY400
                target.log.debug("", exc_info=e)
    except TargetError as e:
        log.error(e)  # noqa: TRY400
        log.debug("", exc_info=e)
        return 1

    return 0


def get_target_info(target: Target, recursive: bool = False) -> dict[str, str | list[str]]:
    return {
        "disks": get_disks_info(target),
        "volumes": get_volumes_info(target),
        "mounts": get_mounts_info(target),
        "children": get_children_info(target, recursive),
        "hostname": get_property(target, "hostname"),
        "domain": get_property(target, "domain"),
        "ips": get_property(target, "ips"),
        "os_family": get_property(target, "os"),
        "os_version": get_property(target, "version"),
        "architecture": get_property(target, "architecture"),
        "language": get_property(target, "language"),
        "timezone": get_property(target, "timezone"),
        "install_date": get_property(target, "install_date"),
        "last_activity": get_property(target, "activity"),
    }


def get_property(target: Target, func: str) -> str | None:
    try:
        if target.has_function(func):
            return getattr(target, func)
    except Exception as e:
        log.warning("Error executing %s: %s", func, e)
        log.debug("", exc_info=e)
    return None


def print_target_info(target: Target, target_info: dict[str, str | list[str]]) -> None:
    print(target)

    for name, value in target_info.items():
        if name in ["disks", "volumes", "mounts", "children"]:
            if not any(value):
                continue
            print(f"\n{name.capitalize()}")
            for i in value:
                values = " ".join([f"{k}={v!r}" for k, v in i.items()])
                print(f"- <{name.capitalize()[:-1].replace('re', '')} {values}>")
            continue

        if isinstance(value, list):
            value = ", ".join(map(str, value))

        if isinstance(value, datetime):
            value = value.isoformat(timespec="microseconds")

        if name == "hostname":
            print()

        print(f"{name.capitalize().replace('_', ' '):14s} : {value}")


def catch_errors(func: Callable[[Target], list[dict]]) -> Callable[[Target], list[dict]]:
    """Catch all errors and return dict with error if encountered."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> list[dict[str, Any]]:
        try:
            return func(*args, **kwargs)
        except Exception as e:
            log.warning("Error executing %s: %s", func.__name__, e)
            log.debug("", exc_info=e)
            return [{"error": str(e)}]

    return wrapper


@catch_errors
def get_disks_info(target: Target) -> list[dict[str, str | int]]:
    return [{"type": d.__type__, "size": d.size} for d in target.disks]


@catch_errors
def get_volumes_info(target: Target) -> list[dict[str, str | int | None]]:
    return [{"name": v.name, "size": v.size, "fs": v.fs.__type__ if v.fs else None} for v in target.volumes]


@catch_errors
def get_mounts_info(target: Target) -> list[dict[str, str | None]]:
    return [{"fs": fs.__type__, "path": path} for path, fs in target.fs.mounts.items()]


@catch_errors
def get_children_info(target: Target, recursive: bool = False) -> list[dict[str, str]]:
    return [
        {"id": child_id, "type": child.type, "name": child.name, "path": str(child.path)}
        for child_id, child in target.list_children(recursive=recursive)
    ]


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import export
from dissect.target.plugins.apps.browser.browser import (
    GENERIC_COOKIE_FIELDS,
    GENERIC_DOWNLOAD_RECORD_FIELDS,
    GENERIC_EXTENSION_RECORD_FIELDS,
    GENERIC_HISTORY_RECORD_FIELDS,
    GENERIC_PASSWORD_RECORD_FIELDS,
    BrowserPlugin,
)
from dissect.target.plugins.apps.browser.chromium import (
    CHROMIUM_DOWNLOAD_RECORD_FIELDS,
    ChromiumMixin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator


class EdgePlugin(ChromiumMixin, BrowserPlugin):
    """Edge browser plugin."""

    __namespace__ = "edge"

    DIRS = (
        # Linux
        ".config/microsoft-edge/Default/",
        ".config/microsoft-edge/Profile*",
        ".var/app/com.microsoft.Edge/config/microsoft-edge/Default",
        ".var/app/com.microsoft.Edge/config/microsoft-edge/Profile*",
        # Windows
        "AppData/Local/Microsoft/Edge/User Data/Default",
        "AppData/Local/Microsoft/Edge/User Data/Profile*",
        "AppData/Local/Microsoft/Edge/User Data/Snapshots/*/Default",
        "AppData/Local/Microsoft/Edge/User Data/Snapshots/*/Profile*",
        # Macos
        "Library/Application Support/Microsoft Edge/Default",
        "Library/Application Support/Microsoft Edge/Profile*",
    )

    BrowserHistoryRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/edge/history", GENERIC_HISTORY_RECORD_FIELDS
    )

    BrowserCookieRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/edge/cookie", GENERIC_COOKIE_FIELDS
    )

    BrowserDownloadRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/edge/download", GENERIC_DOWNLOAD_RECORD_FIELDS + CHROMIUM_DOWNLOAD_RECORD_FIELDS
    )

    BrowserExtensionRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/edge/extension", GENERIC_EXTENSION_RECORD_FIELDS
    )

    BrowserPasswordRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/edge/password", GENERIC_PASSWORD_RECORD_FIELDS
    )

    @export(record=BrowserHistoryRecord)
    def history(self) -> Iterator[BrowserHistoryRecord]:
        """Return browser history records for Microsoft Edge."""
        yield from super().history("edge")

    @export(record=BrowserCookieRecord)
    def cookies(self) -> Iterator[BrowserCookieRecord]:
        """Return browser cookie records for Microsoft Edge."""
        yield from super().cookies("edge")

    @export(record=BrowserDownloadRecord)
    def downloads(self) -> Iterator[BrowserDownloadRecord]:
        """Return browser download records for Microsoft Edge."""
        yield from super().downloads("edge")

    @export(record=BrowserExtensionRecord)
    def extensions(self) -> Iterator[BrowserExtensionRecord]:
        """Return browser extension records for Microsoft Edge."""
        yield from super().extensions("edge")

    @export(record=BrowserPasswordRecord)
    def passwords(self) -> Iterator[BrowserPasswordRecord]:
        """Return browser password records for Microsoft Edge."""
        yield from super().passwords("edge")



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

import pytest

from dissect.target.helpers.regutil import VirtualHive, VirtualKey
from dissect.target.plugins.os.windows.regf.shellbags import (
    ShellBagsPlugin,
    parse_shell_item_list,
)

if TYPE_CHECKING:
    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("shellbag", "name", "modification_time", "localized_name"),
    [
        (
            (
                b"X\x001\x00\x00\x00\x00\x00WX\xc1I\x11\x00MENUST~1\x00\x00@\x00\x03\x00\x04\x00\xef\xbeWXZBWXZB\x14"
                b"\x00*\x00M\x00e\x00n\x00u\x00 \x00S\x00t\x00a\x00r\x00t\x00\x00\x00@shell32.dll,-21786\x00\x18\x00"
                b"\x00\x00"
            ),
            "Menu Start",
            datetime.datetime(2024, 2, 23, 9, 14, 2, tzinfo=datetime.timezone.utc),
            b"@shell32.dll,-21786",
        ),
        (
            (
                b"x\x001\x00\x00\x00\x00\x00\x17W\x0bk\x11\x00Users\x00d\x00\t\x00\x04\x00\xef\xbe\xa7T,*\x91X\xe6R."
                b"\x00\x00\x00\n\x08\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00:\x00\x00\x00\x00\x00A"
                b"\x06\x85\x00U\x00s\x00e\x00r\x00s\x00\x00\x00@\x00s\x00h\x00e\x00l\x00l\x003\x002\x00.\x00d\x00l"
                b"\x00l\x00,\x00-\x002\x001\x008\x001\x003\x00\x00\x00\x14\x00\x00\x00"
            ),
            "Users",
            datetime.datetime(2023, 8, 23, 13, 24, 22, tzinfo=datetime.timezone.utc),
            "@shell32.dll,-21813",
        ),
    ],
    ids=("char", "wchar"),
)
def test_shellbags_parser(
    shellbag: bytes, name: str, modification_time: datetime.datetime, localized_name: str | bytes
) -> None:
    bag = next(parse_shell_item_list(shellbag))

    assert bag.name == name
    assert bag.modification_time == modification_time

    extension = bag.extensions[0]

    assert extension.long_name == name
    assert extension.localized_name == localized_name


@pytest.mark.parametrize(
    ("bags", "expected_type", "expected_path"),
    [
        # single VOLUME with path `A:\`
        (
            [("", "1", "19002f413a5c000000000000000000000000000000000000000000")],
            ["VOLUME"],
            ["A:\\"],
        ),
        # single VOLUME with invalid unicode character
        (
            [("", "0", "19002f4d7920436f6d70757465725c52c3a9ea6d79000000000000")],
            ["VOLUME"],
            ["My Computer\\R\u00e9\udceamy"],
        ),
        # nested path leading to "My Computer\\C:\\Users\\Administrator\\Downloads"
        (
            [
                # ROOT_FOLDER
                ("", "1", "14001f50e04fd020ea3a6910a2d808002b30309d0000"),
                # VOLUME
                ("1", "0", "19002f433a5c000000000000000000000000000000000000000000"),
                # FILE_ENTRY
                (
                    "1\\0",
                    "0",
                    (
                        "7800310000000000875747ab1100557365727300640009000400efbe2f4d2e31"
                        "875747ab2e000000d00100000000010000000000000000003a00000000004899"
                        "e90055007300650072007300000040007300680065006c006c00330032002e00"
                        "64006c006c002c002d0032003100380031003300000014000000"
                    ),
                ),
                # FILE_ENTRY
                (
                    "1\\0\\0",
                    "0",
                    (
                        "6400310000000000875748ab100041444d494e497e3100004c0009000400efbe"
                        "875747ab875748ab2e000000b9f5010000000200000000000000000000000000"
                        "0000688f5a00410064006d0069006e006900730074007200610074006f007200"
                        "000018000000"
                    ),
                ),
                # FILE_ENTRY
                (
                    "1\\0\\0\\0",
                    "0",
                    (
                        "8400310000000000875748ab1100444f574e4c4f7e3100006c0009000400efbe"
                        "875747ab875748ab2e000000c1f5010000000200000000000000000042000000"
                        "0000192d580044006f0077006e006c006f006100640073000000400073006800"
                        "65006c006c00330032002e0064006c006c002c002d0032003100370039003800"
                        "000018000000"
                    ),
                ),
            ],
            ["ROOT_FOLDER", "VOLUME", "FILE_ENTRY", "FILE_ENTRY", "FILE_ENTRY"],
            [
                "My Computer",
                "My Computer\\C:",
                "My Computer\\C:\\Users",
                "My Computer\\C:\\Users\\Administrator",
                "My Computer\\C:\\Users\\Administrator\\Downloads",
            ],
        ),
    ],
    ids=(
        "single-volume",
        "single-volume-invalid-unicode",
        "nested-path",
    ),
)
def test_shellbags_plugin(
    target_win_users: Target,
    hive_hku: VirtualHive,
    bags: list[tuple[str, str, str]],
    expected_type: list[str],
    expected_path: list[str],
) -> None:
    """Test if shellbags mapped to a registry hive are found and parsed correctly."""

    key_name = "Software\\Microsoft\\Windows\\Shell\\BagMRU"

    for bag_key, bag_name, bag_value in bags:
        key = VirtualKey(hive_hku, f"{key_name}\\{bag_key}")
        key.add_value(bag_name, bytes.fromhex(bag_value))
        hive_hku.map_key(f"{key_name}\\{bag_key}", key)

    target_win_users.add_plugin(ShellBagsPlugin)
    results = list(target_win_users.shellbags())

    assert len(results) == len(bags)
    assert [r.type for r in results] == expected_type
    assert [r.path for r in results] == expected_path



from __future__ import annotations

import stat
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export
from dissect.target.plugins.filesystem.walkfs import FilesystemRecord

if TYPE_CHECKING:
    from collections.abc import Iterator

SuidRecord = TargetRecordDescriptor(
    "filesystem/unix/suid",
    FilesystemRecord.target_fields,
)


class SuidPlugin(Plugin):
    """Unix SUID binary plugin."""

    def check_compatible(self) -> None:
        if not self.target.has_function("walkfs") or self.target.os == "windows":
            raise UnsupportedPluginError("Unsupported plugin")

    @export(record=SuidRecord)
    def suid_binaries(self) -> Iterator[SuidRecord]:
        """Return all SUID binaries.

        A SUID binary allows all users to run it with the permissions of its owner. This means that a
        SUID binary owned by the root user can be run with root privileges by any user. Such binaries can be leveraged
        by an adversary to perform privilege escalation.

        References:
            - https://steflan-security.com/linux-privilege-escalation-suid-binaries/
        """
        for record in self.target.walkfs():
            if record.mode & stat.S_ISUID:
                yield SuidRecord(
                    **record._asdict(),
                    _target=self.target,
                )



from __future__ import annotations

import struct
from itertools import chain
from typing import TYPE_CHECKING

from dissect.util.sid import read_sid
from dissect.util.ts import from_unix

from dissect.target.exceptions import RegistryError, UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import (
    RegistryRecordDescriptorExtension,
    UserRecordDescriptorExtension,
)
from dissect.target.helpers.record import (
    TargetRecordDescriptor,
    create_extended_descriptor,
)
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

UserRegistryRecordDescriptor = create_extended_descriptor(
    [
        RegistryRecordDescriptorExtension,
        UserRecordDescriptorExtension,
    ]
)

AppInitRecord = UserRegistryRecordDescriptor(
    "filesystem/registry/appinit",
    [
        ("datetime", "ts"),
        ("path", "path"),
    ],
)

KnownDllRecord = UserRegistryRecordDescriptor(
    "filesystem/registry/knowndlls",
    [
        ("datetime", "ts"),
        ("path", "path"),
    ],
)

SessionManagerRecord = UserRegistryRecordDescriptor(
    "filesystem/registry/sessionmanager",
    [
        ("datetime", "ts"),
        ("path", "path"),
    ],
)

NullSessionPipeRecord = UserRegistryRecordDescriptor(
    "filesystem/registry/nullsessionpipes",
    [
        ("string", "name"),
    ],
)

NdisRecord = UserRegistryRecordDescriptor(
    "filesystem/registry/ndis",
    [
        ("datetime", "ts"),
        ("string", "network_name"),
        ("string", "name"),
        ("string", "pnpinstanceid"),
    ],
)

CommandProcAutoRunRecord = UserRegistryRecordDescriptor(
    "filesystem/registry/commandprocautorun",
    [
        ("datetime", "ts"),
        ("path", "path"),
    ],
)

AlternateShellRecord = UserRegistryRecordDescriptor(
    "filesystem/registry/alternateshell",
    [
        ("datetime", "ts"),
        ("path", "path"),
    ],
)

BootShellRecord = UserRegistryRecordDescriptor(
    "filesystem/registry/bootshell",
    [
        ("datetime", "ts"),
        ("path", "path"),
    ],
)

FileRenameOperationRecord = UserRegistryRecordDescriptor(
    "filesystem/registry/filerenameoperations",
    [
        ("datetime", "ts"),
        ("path", "path"),
    ],
)

WinSockNamespaceProviderRecord = UserRegistryRecordDescriptor(
    "filesystem/registry/winsocknamespaceprovider",
    [
        ("datetime", "ts"),
        ("path", "librarypath"),
        ("string", "displaystring"),
        ("bytes", "providerid"),
        ("boolean", "enabled"),
        ("string", "version"),
    ],
)

ComputerSidRecord = TargetRecordDescriptor(
    "windows/sid/computer",
    [
        ("datetime", "ts"),
        ("string", "sidtype"),
        ("string", "sid"),
    ],
)


class GenericPlugin(Plugin):
    """Generic Windows plugin.

    Provides Windows operating system plugins too small to fit in a separate plugin.
    """

    def check_compatible(self) -> None:
        if not self.target.has_function("registry"):
            raise UnsupportedPluginError("Unsupported Plugin")

    @export(property=True)
    def ntversion(self) -> str | None:
        """Return the Windows NT version."""
        return self.target._os._nt_version()

    @export(output="yield")
    def pathenvironment(self) -> Iterator[str]:
        """Return the content of the Windows PATH environment variable.

        PATH is an environment variable on an operating system that specifies a set of directories where executable
        programs are located. Adversaries may add the directories in which they have stored their (malicious) binaries.

        References:
            - https://en.wikipedia.org/wiki/PATH_%28variable%29
        """
        key = "HKLM\\SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment"
        for r in self.target.registry.keys(key):
            yield r.value("Path").value

    @export(property=True)
    def domain(self) -> str | None:
        """Return the domain name.

        Corporate Windows systems are usually connected to a domain (active directory).

        References:
            - https://en.wikipedia.org/wiki/Active_Directory
        """
        keys = [
            ("HKLM\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Group Policy\\History", "MachineDomain"),
            ("HKLM\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Group Policy\\History", "NetworkName"),
            ("HKLM\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Group Policy\\History", "DCName"),
            ("HKLM\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Telephony", "DomainName"),
        ]

        for key, value in keys:
            try:
                val = self.target.registry.key(key).value(value).value
                if val:
                    return val.strip("\\")
            except RegistryError:  # noqa: PERF203
                continue
        return None

    @export(property=True)
    def activity(self) -> datetime | None:
        """Return last seen activity based on filesystem timestamps."""
        last_seen = 0

        try:
            for f in self.target.resolve("%windir%/system32/winevt/logs").iterdir():
                if f.stat().st_mtime > last_seen:
                    last_seen = f.stat().st_mtime
        except Exception as e:
            self.target.log.debug("Could not determine last activity", exc_info=e)

        try:
            for f in self.target.resolve("%windir%/system32/config").iterdir():
                if f.stat().st_mtime > last_seen:
                    last_seen = f.stat().st_mtime
        except Exception as e:
            self.target.log.debug("Could not determine last activity", exc_info=e)

        if last_seen == 0:
            return None

        return from_unix(last_seen)

    @export(property=True)
    def install_date(self) -> datetime | None:
        """Returns the install date of the system.

        The value of the registry key is stored as a Unix epoch timestamp.

        References:
            - https://winreg-kb.readthedocs.io/en/latest/_modules/winregrc/sysinfo.html?highlight=_ParseInstallDate
            - https://www.forensics-matters.com/2018/09/15/find-out-windows-installation-date/
        """

        key = "HKLM\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion"

        try:
            return from_unix(self.target.registry.key(key).value("InstallDate").value)
        except RegistryError:
            return None

    @export(record=AppInitRecord)
    def appinit(self) -> Iterator[AppInitRecord]:
        """Return all available Application Initial (AppInit) DLLs registry key values.

        AppInit_DLLs is a mechanism that allows an arbitrary list of DLLs to be loaded into each user mode process on
        the system. It can be used as a persistence mechanism and/or elevate privileges by executing malicious content
        triggered by AppInit DLLs loaded into processes. DLLs that are specified in the AppInit_DLLs value in the
        Registry keys HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows NT\\CurrentVersion\\Windows or
        HKEY_LOCAL_MACHINE\\Software\\Wow6432Node\\Microsoft\\Windows NT\\CurrentVersion\\Windows are loaded by
        user32.dll into every process that loads user32.dll.

        References:
            - https://attack.mitre.org/techniques/T1546/010/
            - https://docs.microsoft.com/en-us/windows/win32/win7appqual/appinit-dlls-in-windows-7-and-windows-server-2008-r2?redirectedfrom=MSDN
            - https://docs.microsoft.com/en-US/windows/win32/dlls/secure-boot-and-appinit-dlls
        """
        keys = [
            (
                "HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows NT\\CurrentVersion\\Windows",
                "AppInit_DLLs",
            ),
            (
                "HKEY_LOCAL_MACHINE\\Software\\Wow6432Node\\Microsoft\\Windows NT\\CurrentVersion\\Windows",
                "AppInit_DLLs",
            ),
            ("HKEY_CURRENT_USER\\Software\\Microsoft\\Windows NT\\CurrentVersion\\Windows", "AppInit_DLLs"),
            (
                "HKEY_CURRENT_USER\\Software\\Wow6432Node\\Microsoft\\Windows NT\\CurrentVersion\\Windows",
                "AppInit_DLLs",
            ),
        ]

        for key, name in keys:
            for r in self.target.registry.keys(key):
                user = self.target.registry.get_user(r)
                try:
                    value = r.value(name)
                    yield AppInitRecord(
                        ts=r.ts,
                        path=self.target.fs.path(value.value),
                        _target=self.target,
                        _user=user,
                        _key=r,
                    )
                except RegistryError:
                    continue

    @export(record=KnownDllRecord)
    def knowndlls(self) -> Iterator[KnownDllRecord]:
        """Return all available KnownDLLs registry key values.

        The KnownDLLs registry key values are used to cache frequently used system DLLs. Initially, it was added to
        accelerate application loading, but also it can be considered as a security mechanism, as it prevents malware
        from putting Trojan versions of system DLLs to the application folders (as all main DLLs belong to KnownDLLs,
        the version from the application folder will be ignored). However, these registry keys can still be leveraged
        to perform DLL injection.

        References:
            - https://www.apriorit.com/dev-blog/257-dll-injection
        """
        key = "HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\Control\\Session Manager\\KnownDlls"

        try:
            for r in self.target.registry.keys(key):
                user = self.target.registry.get_user(r)
                for value in r.values():
                    yield KnownDllRecord(
                        ts=r.ts,
                        path=self.target.fs.path(value.value),
                        _target=self.target,
                        _user=user,
                        _key=r,
                    )
        except RegistryError:
            pass

    @export(record=SessionManagerRecord)
    def sessionmanager(self) -> Iterator[SessionManagerRecord]:
        """Return interesting Session Manager (Smss.exe) registry key entries.

        Session Manager (Smss.exe) is the first user-mode process started by the kernel and performs several tasks,
        such as creating environment variables, starts the Windows Logon Manager (winlogon.exe), etc. The BootExecute
        registry key holds the Windows tasks that cannot be performed when Windows is running, the Execute registry key
        should never be populated when Windows is installed. Can be leveraged as persistence mechanisms.

        References:
            - https://en.wikipedia.org/wiki/Session_Manager_Subsystem
            - https://www.microsoftpressstore.com/articles/article.aspx?p=2762082&seqNum=2
        """
        keys = [
            ("HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\Control\\Session Manager", "BootExecute"),
            ("HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\Control\\Session Manager", "Execute"),
            ("HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\Control\\Session Manager\\SubSystems", "windows"),
            ("HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\Control\\Session Manager\\WOW", "cmdline"),
            ("HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\Control\\Session Manager\\WOW", "wowcmdline"),
        ]

        for key, name in keys:
            for r in self.target.registry.keys(key):
                try:
                    value = r.value(name)
                    data = value.value
                except RegistryError:
                    continue

                user = self.target.registry.get_user(r)

                if isinstance(data, list):
                    for d in data:
                        if d == "autocheck autochk *":
                            continue

                        yield SessionManagerRecord(
                            ts=r.ts,
                            path=self.target.fs.path(d),
                            _target=self.target,
                            _user=user,
                            _key=r,
                        )
                else:
                    yield SessionManagerRecord(
                        ts=r.ts,
                        path=self.target.fs.path(data.split(" ")[0]),
                        _target=self.target,
                        _user=user,
                        _key=r,
                    )

    @export(record=NullSessionPipeRecord)
    def nullsessionpipes(self) -> Iterator[NullSessionPipeRecord]:
        """Return the NullSessionPipes registry key value.

        The NullSessionPipes registry key value specifies server pipes and shared folders that are excluded from the
        policy that does not allow null session access. A null session implies that access to a network resource, most
        commonly the IPC$ "Windows Named Pipe" share, was granted without authentication. Also known as anonymous or
        guest access. These can thus be accessed without authentication and can be leveraged for latteral movement
        and/or privilege escalation.

        References:
            - https://docs.microsoft.com/en-us/windows/security/threat-protection/security-policy-settings/network-access-restrict-anonymous-access-to-named-pipes-and-shares
        """
        key = "HKLM\\System\\CurrentControlSet\\Services\\LanManServer\\Parameters"
        for r in self.target.registry.keys(key):
            user = self.target.registry.get_user(r)
            try:
                value = r.value("NullSessionPipes")
                for pipe in value.value:
                    yield NullSessionPipeRecord(
                        name=pipe,
                        _target=self.target,
                        _user=user,
                        _key=r,
                    )
            except RegistryError:
                continue

    @export(record=NdisRecord)
    def ndis(self) -> Iterator[NdisRecord]:
        """Return network registry key entries."""
        key = "HKLM\\System\\CurrentControlSet\\Control\\Network\\{4D36E972-E325-11CE-BFC1-08002BE10318}"
        for r in self.target.registry.keys(key):
            user = self.target.registry.get_user(r)
            for sub in r.subkeys():
                for network in sub.subkeys():
                    if network.name == "Descriptions":
                        continue

                    name = None
                    pnpinstanceid = None

                    try:
                        name = network.value("Name").value
                    except RegistryError:
                        pass

                    try:
                        pnpinstanceid = network.value("PnpInstanceID").value
                    except RegistryError:
                        pass

                    yield NdisRecord(
                        ts=network.ts,
                        network_name=sub.name,
                        name=name,
                        pnpinstanceid=pnpinstanceid,
                        _target=self.target,
                        _user=user,
                        _key=network,
                    )

    @export(record=CommandProcAutoRunRecord)
    def commandprocautorun(self) -> Iterator[CommandProcAutoRunRecord]:
        """Return all available Command Processor (cmd.exe) AutoRun registry key values.

        The Command Processor AutoRun registry key values contain commands that are run each time the Command Processor
        (cmd.exe) is started. Since these commands are not shown to the user in the Command Processor, it can be
        exploited by an adversary to hide malicious commands or leverage as a persistence mechanism

        References:
            - https://devblogs.microsoft.com/oldnewthing/20071121-00/?p=24433
            - https://docs.microsoft.com/en-us/previous-versions/windows/it-pro/windows-server-2003/cc779439%28v=ws.10%29?redirectedfrom=MSDN
        """
        keys = [
            ("HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Command Processor", "AutoRun"),
            ("HKEY_LOCAL_MACHINE\\Software\\Wow6432Node\\Microsoft\\Command Processor", "AutoRun"),
            ("HKEY_CURRENT_USER\\Software\\Microsoft\\Command Processor", "AutoRun"),
            ("HKEY_CURRENT_USER\\Software\\Wow6432Node\\Microsoft\\Command Processor", "AutoRun"),
        ]

        for key, name in keys:
            for r in self.target.registry.keys(key):
                user = self.target.registry.get_user(r)
                try:
                    value = r.value(name)
                    yield CommandProcAutoRunRecord(
                        ts=r.ts,
                        path=self.target.fs.path(value.value),
                        _target=self.target,
                        _user=user,
                        _key=r,
                    )
                except RegistryError:
                    continue

    @export(record=AlternateShellRecord)
    def alternateshell(self) -> Iterator[AlternateShellRecord]:
        """Return the AlternateShell registry key value.

        The AlternateShell registry key, HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\Control\\Safeboot, specifies the
        shell that is used when a Windows system is started in "Safe Mode with Command Prompt". Can be leveraged as a
        persistence mechanism.

        References:
            - https://technet.microsoft.com/en-us/library/cc976124.aspx
        """
        key = "HKEY_LOCAL_MACHINE\\System\\CurrentControlSet\\Control\\Safeboot"

        for r in self.target.registry.keys(key):
            user = self.target.registry.get_user(r)
            value = r.value("AlternateShell")
            yield AlternateShellRecord(
                ts=r.ts,
                path=self.target.fs.path(value.value),
                _target=self.target,
                _user=user,
                _key=r,
            )

    @export(record=BootShellRecord)
    def bootshell(self) -> Iterator[BootShellRecord]:
        """Return the BootShell registry key entry.

        Usually contains a path to bootim.exe which is Windows's recovery menu.
        This registry key can be used as a persistence mechanism.
        """
        key = "HKLM\\SYSTEM\\CurrentControlSet\\Control\\Session Manager"

        for r in self.target.registry.keys(key):
            user = self.target.registry.get_user(r)
            try:
                value = r.value("BootShell")
            except RegistryError:
                continue

            yield BootShellRecord(
                ts=r.ts,
                path=self.target.fs.path(value.value),
                _target=self.target,
                _user=user,
                _key=r,
            )

    @export(record=FileRenameOperationRecord)
    def filerenameop(self) -> Iterator[FileRenameOperationRecord]:
        """Return all pending file rename operations.

        The PendingFileRenameOperations registry key value contains information about files that will be renamed on
        reboot. Can be used to hunt for malicious binaries.

        References:
            - https://forensicatorj.wordpress.com/2014/06/25/interpreting-the-pendingfilerenameoperations-registry-key-for-forensics/
            - https://docs.microsoft.com/en-us/previous-versions/windows/it-pro/windows-2000-server/cc960241%28v=technet.10%29?redirectedfrom=MSDN
            - https://qtechbabble.wordpress.com/2020/06/26/use-pendingfilerenameoperations-registry-key-to-automatically-delete-a-file-on-reboot/
        """
        key = "HKLM\\System\\CurrentControlSet\\Control\\Session Manager"
        for r in self.target.registry.keys(key):
            user = self.target.registry.get_user(r)
            try:
                value = r.value("PendingFileRenameOperations")
                paths = map(self.target.fs.path, value.value)
            except RegistryError:
                continue

            for file_path in paths:
                yield FileRenameOperationRecord(
                    ts=r.ts,
                    path=file_path,
                    _target=self.target,
                    _user=user,
                    _key=r,
                )

    @export(record=WinSockNamespaceProviderRecord)
    def winsocknamespaceprovider(self) -> Iterator[WinSockNamespaceProviderRecord]:
        """Return available protocols stored in the Winsock catalog database.

        References:
            - https://docs.microsoft.com/en-us/windows/win32/winsock/name-space-service-providers-2?redirectedfrom=MSDN
        """
        keys = [
            "HKLM\\System\\CurrentControlSet\\Services\\Winsock2\\Parameters\\namespace_catalog5\\catalog_entries",
            "HKLM\\System\\CurrentControlSet\\Services\\Winsock2\\Parameters\\namespace_catalog5\\catalog_entries64",
        ]

        for key in keys:
            for r in self.target.registry.keys(key):
                user = self.target.registry.get_user(r)
                for s in r.subkeys():
                    yield WinSockNamespaceProviderRecord(
                        ts=r.ts,
                        librarypath=self.target.fs.path(s.value("LibraryPath").value),
                        displaystring=s.value("DisplayString").value,
                        providerid=s.value("ProviderID").value,
                        enabled=s.value("Enabled").value,
                        version=s.value("Version").value,
                        _target=self.target,
                        _user=user,
                        _key=s,
                    )

    @export(property=True)
    def codepage(self) -> str | None:
        """Returns the current active codepage on the system."""

        key = "HKLM\\SYSTEM\\CurrentControlSet\\Control\\Nls\\CodePage"

        try:
            return self.target.registry.key(key).value("ACP").value
        except RegistryError:
            pass

    @export(record=ComputerSidRecord)
    def machine_sid(self) -> Iterator[ComputerSidRecord]:
        """Return the machine SID of the system."""

        try:
            key = self.target.registry.key("HKLM\\SAM\\SAM\\Domains\\Account")

            # The machine SID is stored in the last 12 bytes of the V value as little-endian
            # The machine SID differs from a 'normal' binary SID as it only holds 3 values and lacks a prefix / Revision
            # NOTE: Consider moving this to dissect.util.sid if we encounter this more often
            sid = struct.unpack_from("<III", key.value("V").value, -12)
            yield ComputerSidRecord(
                ts=key.timestamp,
                sidtype="Machine",
                sid=f"S-1-5-21-{sid[0]}-{sid[1]}-{sid[2]}",
                _target=self.target,
            )
        except (RegistryError, struct.error) as e:
            self.target.log.error("Cannot read machine SID from registry")  # noqa: TRY400
            self.target.log.debug("", exc_info=e)
            return None

    @export(record=ComputerSidRecord)
    def domain_sid(self) -> Iterator[ComputerSidRecord]:
        """Return the domain SID of the system.

        The domain SID is stored in the registry under HKLM\\SECURITY\\Policy\\PolMachineAccountS.
        """

        try:
            key = self.target.registry.key("HKLM\\SECURITY\\Policy\\PolMachineAccountS")
            raw_sid = key.value("(Default)").value

            yield ComputerSidRecord(
                ts=key.timestamp,
                sidtype="Domain",
                sid=read_sid(raw_sid) if raw_sid else None,
                _target=self.target,
            )
        except (RegistryError, struct.error):
            self.target.log.exception("Cannot read domain SID from registry")
            return None

    @export(record=ComputerSidRecord)
    def sid(self) -> Iterator[ComputerSidRecord]:
        """Return the machine- and optional domain SID of the system."""

        yield from chain(self.machine_sid(), self.domain_sid())



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.plugins.apps.chat.msn import MSNPlugin, convert_email
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_msn(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    """Test if we parse MSN Chat messages on Windows XP correctly."""

    morpheus_id = convert_email("morpheus@matrix.internal")
    neo_id = convert_email("neo@matrix.internal")

    assert morpheus_id == 2450688751
    assert neo_id == 4092013818

    fs_win.makedirs(f"Users/John/Application Data/Microsoft/MSN Messenger/{morpheus_id}")
    fs_win.map_file(
        f"Users/John/My Documents/My Received Files/morpheus{morpheus_id}/History/neo{neo_id}.xml",
        absolute_path("_data/plugins/apps/chat/msn/history.xml"),
    )

    target_win_users.add_plugin(MSNPlugin)
    assert len(target_win_users.msn.installs) == 1

    results = list(target_win_users.msn.history())
    assert len(results) == 35

    assert results[0].username == "John"
    assert results[0].hostname is None

    assert results[0].ts == datetime(2025, 4, 1, 13, 37, 0, tzinfo=timezone.utc)
    assert results[0].client == "msn"
    assert results[0].account == str(morpheus_id)
    assert results[0].sender == "morpheus@matrix.internal"
    assert results[0].recipient == "neo@matrix.internal"

    assert [(r.sender.replace("@matrix.internal", ""), r.message) for r in results] == [
        (
            "morpheus",
            "At last.",
        ),
        (
            "morpheus",
            "Welcome, Neo. As you no doubt have guessed, I am Morpheus.",
        ),
        (
            "neo",
            "It's an honor.",
        ),
        (
            "morpheus",
            "No, the honor is mine. Please. Come. Sit.",
        ),
        (
            "morpheus",
            "I imagine, right now, you must be feeling a bit like Alice, tumbling down the rabbit hole?",
        ),
        (
            "neo",
            "You could say that.",
        ),
        (
            "morpheus",
            "I can see it in your eyes. You have the look of a man who accepts "
            "what he sees because he is expecting to wake up.",
        ),
        (
            "morpheus",
            "Ironically, this is not far from the truth. But I'm getting ahead of "
            "myself. Can you tell me, Neo, why are you here?",
        ),
        (
            "neo",
            "You're Morpheus. You're a legend. Most hackers would die to meet you.",
        ),
        (
            "morpheus",
            "Yes. Thank you. But I think we both know there's more to it than that. Do you believe in fate, Neo?",
        ),
        (
            "neo",
            "No.",
        ),
        (
            "morpheus",
            "Why not?",
        ),
        (
            "neo",
            "Because I don't like the idea that I'm not in control of my life.",
        ),
        (
            "morpheus",
            "I know exactly what you mean.",
        ),
        (
            "morpheus",
            "Let me tell you why you are here. You have come because you know something.",
        ),
        (
            "morpheus",
            "What you know you can't explain but you feel it.",
        ),
        (
            "morpheus",
            "You've felt it your whole life, felt that something is wrong with the world.",
        ),
        (
            "morpheus",
            "You don't know what, but it's there like a splinter in your mind, "
            "driving you mad. It is this feeling that brought you to me.",
        ),
        (
            "morpheus",
            "Do you know what I'm talking about?",
        ),
        (
            "neo",
            "The Matrix?",
        ),
        (
            "morpheus",
            "Do you want to know what it is?",
        ),
        (
            "morpheus",
            "The Matrix is everywhere, it's all around us, here even in this room.",
        ),
        (
            "morpheus",
            "You can see it out your window or on your television. You feel it "
            "when you go to work, or go to church or pay your taxes.",
        ),
        (
            "morpheus",
            "It is the world that has been pulled over your eyes to blind you from the truth.",
        ),
        (
            "neo",
            "What truth?",
        ),
        (
            "morpheus",
            "That you are a slave, Neo. Like everyone else, you were born into "
            "bondage, kept inside a prison that you cannot smell, taste, or touch.",
        ),
        (
            "morpheus",
            "A prison for your mind.",
        ),
        (
            "morpheus",
            "Unfortunately, no one can be told what the Matrix is.",
        ),
        (
            "morpheus",
            "You have to see it for yourself.",
        ),
        (
            "morpheus",
            "This is your last chance. After this, there is no going back.",
        ),
        (
            "morpheus",
            "You take the blue pill and the story ends.",
        ),
        (
            "morpheus",
            "You wake in your bed and you believe whatever you want to believe.",
        ),
        (
            "morpheus",
            "You take the red pill and you stay in Wonderland and I show you how deep the rabbit-hole goes.",
        ),
        (
            "morpheus",
            "Remember that all I am offering is the truth. Nothing more.",
        ),
        (
            "morpheus",
            "Follow me.",
        ),
    ]
